#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
ROOTPATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
import sys
sys.path.insert(0, ROOTPATH)
import geoclide as gc
import numpy as np
import math


def test_triangle_intersection():
    p0 = gc.Point(-1., -1., 0.)
    p1 = gc.Point(1., -1., 0.)
    p2 = gc.Point(1., 1., 0.)

    tri = gc.Triangle(p0, p1, p2)
    ray = gc.Ray(o=gc.Point(0., 0., 1.), d=gc.normalize(gc.Vector(0.999,0.999,-1.)))

    thitv2, dgv2, is_intv2 = tri.intersect_v2(ray)
    thitv3, dgv3, is_intv3 = tri.intersect_v3(ray)

    assert (is_intv2 is True), 'Problem with v2 intersection test'
    assert (is_intv3 is True), 'Problem with v3 intersection test'
    assert (np.isclose(1.73089629960896, thitv2, 0., 1e-14)), 'Problem with v2 intersection test'
    assert (np.isclose(1.73089629960896, thitv3, 0., 1e-14)), 'Problem with v3 intersection test'
    assert (dgv2.n == gc.Normal(0., 0., 1.)), 'Problem with v2 intersection test'
    assert (dgv3.n == gc.Normal(0., 0., 1.)), 'Problem with v3 intersection test'

    p3 = gc.Point(0.999, 0.999, 0.)

    assert (np.isclose(dgv2.p.x, p3.x, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv2.p.y, p3.y, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv2.p.z, p3.z, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv3.p.x, p3.x, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(dgv3.p.y, p3.y, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(dgv3.p.z, p3.z, 0., 1e-15)), 'Problem with v3 triangle intersection test'

    # Bellow the ray cannot reach the triangle
    ray = gc.Ray(o=gc.Point(0., 0., 1.), d=gc.normalize(gc.Vector(0.999,0.999,-1.)), maxt=1.7)

    thitv2, dgv2, is_intv2 = tri.intersect_v2(ray)
    thitv3, dgv3, is_intv3 = tri.intersect_v3(ray)
    
    assert (is_intv2 is False), 'Problem with v2 intersection test'
    assert (is_intv3 is False), 'Problem with v3 intersection test'

def test_triangle_transform():
    p0 = gc.Point(-0.5, 0.5, 0.)
    p1 = gc.Point(0.5, 0.5, 0.)
    p2 = gc.Point(0.5, -0.5, 0.)

    oTw = gc.get_translate_tf(gc.Vector(10., 0., 5.)) * gc.get_rotateY_tf(45.)
    tri = gc.Triangle(p0, p1, p2, oTw=oTw)
    ray = gc.Ray(o=gc.Point(0., 0., 4.8), d=gc.normalize(gc.Vector(1.,0.,0.)))

    thitv2, dgv2, is_intv2 = tri.intersect(ray, method='v2')
    thitv3, dgv3, is_intv3 = tri.intersect(ray, method='v3')

    assert (is_intv2 is True), 'Problem with v2 intersection test'
    assert (is_intv3 is True), 'Problem with v3 intersection test'
    assert (np.isclose(10.2, thitv2, 0., 1e-14)), 'Problem with v2 intersection test'
    assert (np.isclose(10.2, thitv3, 0., 1e-14)), 'Problem with v3 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., dgv2.n.x, 0., 1e-13)), \
        'Problem with v2 intersection test'
    assert (dgv2.n.y == 0.), 'Problem with v2 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., dgv2.n.z, 0., 1e-13)), \
        'Problem with v2 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., dgv3.n.x, 0., 1e-13)), \
        'Problem with v3 intersection test'
    assert (dgv3.n.y == 0.), 'Problem with v3 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., dgv3.n.z, 0., 1e-13)), \
        'Problem with v3 intersection test'

    p3 = gc.Point(10.2, 0., 4.8)

    assert (np.isclose(dgv2.p.x, p3.x, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv2.p.y, p3.y, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv2.p.z, p3.z, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv3.p.x, p3.x, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(dgv3.p.y, p3.y, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(dgv3.p.z, p3.z, 0., 1e-15)), 'Problem with v3 triangle intersection test'

    assert (tri.is_intersection(ray, method='v2')), 'Problem with v2 is_intersection test'
    assert (tri.is_intersection(ray, method='v3')), 'Problem with v3 is_intersection test'


def test_triangle_mesh():
    # list of vertices
    v = np.array([ [-0.5, -0.5, 0.],               # v0
                [0.5, -0.5, 0.],                   # v1
                [-0.5, 0.5, 0.],                   # v2
                [0.5, 0.5, 0.]], dtype=np.float64) # v3

    vi = np.array([0, 1, 2,                # vertices index of T0
                2, 3, 1], dtype=np.int32)  # vertices index of T1

    oTw = gc.get_translate_tf(gc.Vector(10., 0., 5.)) * gc.get_rotateY_tf(45.)
    tri_mesh = gc.TriangleMesh(vi=vi, v=v, oTw=oTw)
    ray = gc.Ray(o=gc.Point(0., 0., 4.8), d=gc.normalize(gc.Vector(1.,0.,0.)))

    thitv2, dgv2, is_intv2 = tri_mesh.intersect(ray, method='v2')
    thitv3, dgv3, is_intv3 = tri_mesh.intersect(ray, method='v3')

    assert (is_intv2 is True), 'Problem with v2 intersection test'
    assert (is_intv3 is True), 'Problem with v3 intersection test'
    assert (np.isclose(10.2, thitv2, 0., 1e-14)), 'Problem with v2 intersection test'
    assert (np.isclose(10.2, thitv3, 0., 1e-14)), 'Problem with v3 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., dgv2.n.x, 0., 1e-13)), \
        'Problem with v2 intersection test'
    assert (dgv2.n.y == 0.), 'Problem with v2 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., dgv2.n.z, 0., 1e-13)), \
        'Problem with v2 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., dgv3.n.x, 0., 1e-13)), \
        'Problem with v3 intersection test'
    assert (dgv3.n.y == 0.), 'Problem with v3 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., dgv3.n.z, 0., 1e-13)), \
        'Problem with v3 intersection test'

    p3 = gc.Point(10.2, 0., 4.8)

    assert (np.isclose(dgv2.p.x, p3.x, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv2.p.y, p3.y, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv2.p.z, p3.z, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(dgv3.p.x, p3.x, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(dgv3.p.y, p3.y, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(dgv3.p.z, p3.z, 0., 1e-15)), 'Problem with v3 triangle intersection test'

    assert (tri_mesh.is_intersection(ray, method='v2')), 'Problem with v2 is_intersection test'
    assert (tri_mesh.is_intersection(ray, method='v3')), 'Problem with v3 is_intersection test'


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import math
import os
ROOTPATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
import sys
sys.path.insert(0, ROOTPATH)
import geoclide as gc


def test_dot():
    assert (gc.dot(gc.Vector(0., 0., 1.), gc.Vector(0., 1., 0.)) == 0.)
    assert (gc.dot(gc.Vector(0., 0., 1.), gc.Vector(0., 0., 1.)) == 1.)
    v = gc.Vector(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.dot(gc.Vector(0., 0., 1.), v) == math.sqrt(2.)/2.)

    assert (gc.dot(gc.Normal(0., 0., 1.), gc.Vector(0., 1., 0.)) == 0.)
    assert (gc.dot(gc.Normal(0., 0., 1.), gc.Vector(0., 0., 1.)) == 1.)
    v = gc.Normal(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.dot(gc.Vector(0., 0., 1.), v) == math.sqrt(2.)/2.)

    assert (gc.dot(gc.Vector(0., 0., 1.), gc.Normal(0., 1., 0.)) == 0.)
    assert (gc.dot(gc.Vector(0., 0., 1.), gc.Normal(0., 0., 1.)) == 1.)
    v = gc.Normal(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.dot(gc.Vector(0., 0., 1.), v) == math.sqrt(2.)/2.)

    assert (gc.dot(gc.Normal(0., 0., 1.), gc.Normal(0., 1., 0.)) == 0.)
    assert (gc.dot(gc.Normal(0., 0., 1.), gc.Normal(0., 0., 1.)) == 1.)
    v = gc.Normal(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.dot(gc.Normal(0., 0., 1.), v) == math.sqrt(2.)/2.)


def test_cross():
    assert (gc.cross(gc.Vector(0., 0., 1.), gc.Vector(0., 1., 0.)) == gc.Vector(-1.0, 0.0, 0.0))
    assert (gc.cross(gc.Vector(0., 0., 1.), gc.Vector(0., 0., 1.)) == gc.Vector(0.0, 0.0, 0.0))
    v = gc.Vector(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.cross(gc.Vector(0., 0., 1.), v) ==  gc.Vector(0.0, math.sqrt(2.)/2., 0.0))

    assert (gc.cross(gc.Normal(0., 0., 1.), gc.Vector(0., 1., 0.)) == gc.Vector(-1.0, 0.0, 0.0))
    assert (gc.cross(gc.Normal(0., 0., 1.), gc.Vector(0., 0., 1.)) == gc.Vector(0.0, 0.0, 0.0))
    v = gc.Vector(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.cross(gc.Normal(0., 0., 1.), v) ==  gc.Vector(0.0, math.sqrt(2.)/2., 0.0))

    assert (gc.cross(gc.Vector(0., 0., 1.), gc.Normal(0., 1., 0.)) == gc.Vector(-1.0, 0.0, 0.0))
    assert (gc.cross(gc.Vector(0., 0., 1.), gc.Normal(0., 0., 1.)) == gc.Vector(0.0, 0.0, 0.0))
    v = gc.Normal(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.cross(gc.Vector(0., 0., 1.), v) ==  gc.Vector(0.0, math.sqrt(2.)/2., 0.0))


def test_normalize():
    length = math.sqrt(1. + 4. + 9.)
    a_1 = 1. / length
    a_2 = 2./ length
    a_3 = 3. / length
    assert (gc.normalize(gc.Vector(1.,2.,3.)) == gc.Vector(a_1, a_2, a_3))
    assert (gc.normalize(gc.Vector(-1.,-2.,-3.)) == gc.Vector(-a_1, -a_2, -a_3))

    assert (gc.normalize(gc.Normal(1.,2.,3.)) == gc.Normal(a_1, a_2, a_3))
    assert (gc.normalize(gc.Normal(-1.,-2.,-3.)) == gc.Normal(-a_1, -a_2, -a_3))


def test_coordinate_system():
    v1 = gc.Vector(0., 0., 1.)
    v2_m1, v3_m1 = gc.coordinate_system(v1, 'm1')
    assert (v2_m1 == gc.Vector(0.0, 1.0, -0.0))
    assert (v3_m1 == gc.Vector(-1.0, 0.0, 0.0))
    v2_m2, v3_m2 = gc.coordinate_system(v1, 'm2')
    assert (v2_m2 == gc.Vector(1.0, -0.0, -0.0))
    assert (v3_m2 == gc.Vector(-0.0, 1.0, -0.0))


def test_distance():
    p1 = gc.Point(0., 0., 0.)
    p2 = gc.Point(0., 0., 10.)
    assert (gc.distance(p1,p2) == 10.)
    p1 = gc.Point(1., 2., 1.9)
    p2 = gc.Point(5., 15., 3.)
    a_1 = math.sqrt(16 + 13**2 + 1.1**2)
    assert (gc.distance(p1,p2) == a_1)


def test_face_forward():
    n1 = gc.Normal(1., 0., 0.)
    v1 = gc.Vector(-1., 0., 0.)
    assert (gc.face_forward(v1, n1) == gc.Vector(1.0, -0.0, -0.0))
    n1 = gc.Normal(-1., 0., 0.)
    v1 = gc.Vector(-1., 0., 0.)
    assert (gc.face_forward(v1, n1) == gc.Vector(-1.0, 0.0, 0.0))


def test_vmax():
    v1 = gc.Vector(2.,3.,1.)
    assert (gc.vmax(v1) == 3.)


def test_vmin():
    v1 = gc.Vector(2.,3.,1.)
    assert (gc.vmin(v1) == 1.)


def test_vargmax():
    v1 = gc.Vector(2.,3.,1.)
    assert (gc.vargmax(v1) == 1)


def test_vargmin():
    v1 = gc.Vector(2.,3.,1.)
    assert (gc.vargmin(v1) == 2)


def test_permute():
    v1 = gc.Vector(2., 3., 1.)
    assert (gc.permute(v1, 1, 0, 2) == gc.Vector(3.0, 2.0, 1.0))
    assert (gc.permute(v1, np.array([1, 0, 2])) == gc.Vector(3.0, 2.0, 1.0))





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import math


def clamp(val, val_min, val_max):
    """
    Clamps val into the range [val_min, val_max]

    Parameters
    ----------
    val : float
        The scalar to be clamped
    val_min : float
        The minumum value
    val_max : The maximum value

    Results
    -------
    out : float
        The result of the clamp

    Examples
    --------
    >>> import geoclide as gc
    >>> gc.clamp(4, val_min=5, val_max=11)
    5
    """
    if (not np.isscalar(val)     or 
        not np.isscalar(val_min) or
        not np.isscalar(val_max) ):
        raise ValueError('The parameters must be all scalars')
    
    return val_min if (val < val_min) else (val_max if (val > val_max) else val)


def quadratic(a, b, c):
    """
    Resolve the quadratic polynomial: ax**2 + bx + c

    - where x is the quadratic polynomial variable and a, b and c the coefficients

    Parameters
    ----------
    a : float
        The first coefficient of the quadratic polynomial
    b : float
        The second coefficient of the quadratic polynomial
    c : float
        The third coefficient of the quadratic polynomial

    Results
    -------
    b : bool
        If the quadratic can be solved -> True, else False
    x0 : float | None
        The first solution
    x1 : float | None 
        The second solution

    Notes
    -----
    If There are 2 solutions x0 < x1. And if there is only one solution x0 = x1.

    Examples
    --------
    >>> import geoclide as gc
    >>> a = 2
    >>> b = -5
    >>> c = 0
    >>> gc.quadratic(a, b, c)
    (True, 0.0, 2.5)
    """
    if (not np.isscalar(a) or not np.isscalar(b) or not np.isscalar(c)):
        raise ValueError('The parameters must be all scalars')
    
    #  Find quadratic discriminant
    discrim = (b * b) - (4 * a * c)

    if (discrim < 0): return False, None, None

    rootDiscrim = math.sqrt(discrim)

    # Compute quadratic xi values
    if (b < 0): q = -0.5 * (b - rootDiscrim)
    else: q = -0.5 * (b + rootDiscrim)

    if (a != 0): x0 = q / a
    else: x0 = c / q

    x1 = c / q

    if (x0 > x1): x0, x1 = x1, x0

    return True, x0, x1


def gamma_f32(n):
    """
    Gamma function from pbrt v3
    """
    epsi = np.finfo(np.float32).eps * 0.5
    return (n*epsi)/(1 - n*epsi)


def gamma_f64(n):
    """
    Gamma function from pbrt v3 but in double precision
    """
    epsi = np.finfo(np.float64).eps * 0.5
    return (n*epsi)/(1 - n*epsi)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.shapes import Shape, DifferentialGeometry
from geoclide.mathope import clamp, quadratic
from geoclide.basic import Ray, Vector
from geoclide.transform import Transform
import math
import numpy as np


class Sphere(Shape):
    '''
    Creation of the class Sphere

    - without transformation the sphere is centered at the origin
    - z0, z1 and phi_max are needed parameters for the creation of any partial sphere

    Parameters
    ----------
    radius : float
        The radius of the sphere
    z_min : float, optional
        The minimum z value of the sphere where z0 is between [-radius, 0]
    z_max : float, optional
        The maximum z value of the sphere where z1 is between [0, radius]
    phi_max : float, optional
        The maximum phi value in radians of the sphere, where phi is between [0, 360]
    oTw : Transform, optional
        From object to world space or the transformation applied to the sphere
    wTo : Transform, optional
        From world to object space or the in inverse transformation applied to the sphere
    '''
    def __init__(self, radius, z_min=None, z_max=None, phi_max=2*math.pi, oTw=None, wTo=None):
        if z_min is None: z_min = -radius
        if z_max is None: z_max = radius
        if wTo is None and oTw is None:
            wTo = Transform()
            oTw = Transform()
        elif (wTo is None and isinstance(oTw, Transform)): wTo = oTw.inverse()
        elif (isinstance(wTo, Transform) and oTw is None): oTw = wTo.inverse()
        if (not np.isscalar(radius) or
            not np.isscalar(z_min)  or
            not np.isscalar(z_max)  or
            not np.isscalar(phi_max) ):
            raise ValueError('The parameters radius, z_min, z_max and phi_max must be all scalars')
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        self.radius = radius
        self.zmin = clamp(z_min, -self.radius, self.radius)
        self.zmax = clamp(z_max, -self.radius, self.radius)
        self.thetaMin = math.acos(clamp(self.zmin/self.radius, -1, 1))
        self.thetaMax = math.acos(clamp(self.zmax/self.radius, -1, 1))
        self.phiMax = clamp(phi_max, 0, 2*math.pi)

    def is_intersection(self, r1):
        """
        Test if a ray intersects the sphere / partial sphere

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test

        Returns
        -------
        out : bool
            If there is an intersection -> True, else False

        Examples
        --------
        >>> import geoclide as gc
        >>> sph1 = gc.Sphere(radius=1.) # sphere of radius 1
        >>> sph2 = gc.Sphere(radius=1., z_max=0.5) # partial sphere where portion above z=0.5 is removed
        >>> r = gc.Ray(o=gc.Point(-2., 0., 0.8), d=gc.Vector(1.,0.,0.))
        >>> sph1.is_intersection(r)
        True
        >>> thit, dg, is_int = sph1.intersect(r)
        >>> sph2.intersect(r) # here no intersection since the sphere part above z=0.5 is removed
        False
        """
        if not isinstance(r1, Ray): raise ValueError('The given parameter must be a Ray')
        ray = Ray(r1)
        ray.o = self.wTo[r1.o]
        ray.d = self.wTo[r1.d]

        # Compute quadratic sphere coefficients
        a = ray.d.x*ray.d.x + ray.d.y*ray.d.y + ray.d.z*ray.d.z
        b = 2 * (ray.d.x*ray.o.x + ray.d.y*ray.o.y + ray.d.z*ray.o.z)
        c = ray.o.x*ray.o.x + ray.o.y*ray.o.y + ray.o.z*ray.o.z - \
            self.radius*self.radius

        # Solve quadratic equation
        exist, t0, t1 = quadratic(a, b, c)
        if (not exist): return False
        
        # Compute intersection distance along ray
        if (t0 > ray.maxt or t1 < ray.mint): return False
        thit = t0

        if (t0 < ray.mint):
            thit = t1
            if (thit > ray.maxt): return False

        # Compute sphere hit position and $\phi$
        phit = ray[thit]
        if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
        phi = math.atan2(phit.y, phit.x)
        if (phi < 0): phi += 2*math.pi

        # Test sphere intersection against clipping parameters
        if ((self.zmin > -self.radius and phit.z < self.zmin) or
            (self.zmax <  self.radius and phit.z > self.zmax) or
            (phi > self.phiMax) ):
            if (thit == t1): return False
            if (t1 > ray.maxt): return False
            thit = t1
            # Compute sphere hit position and $\phi$
            phit = ray[thit]
            if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
            phi = math.atan2(phit.y, phit.x)
            if (phi < 0): phi += 2*math.pi
            if ((self.zmin > -self.radius and phit.z < self.zmin) or
                (self.zmax <  self.radius and phit.z > self.zmax) or
                (phi > self.phiMax) ):
                return False

        return True

    def intersect(self, r1):
        """
        Test if a ray intersects the sphere / partial sphere

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test

        Returns
        -------
        thit : float
            The t ray variable for its first intersection at the shape surface
        dg : DifferentialGeometry
            The parametric parameters at the intersection point
        is_intersection : bool
            If there is an intersection -> True, else False

        Examples
        --------
        >>> import geoclide as gc
        >>> sph1 = gc.Sphere(radius=1.) # sphere of radius 1
        >>> sph2 = gc.Sphere(radius=1., z_max=0.5) # partial sphere where portion above z=0.5 is removed
        >>> r = gc.Ray(o=gc.Point(-2., 0., 0.8), d=gc.Vector(1.,0.,0.))
        >>> sph1.intersect(r)
        (19.399999999999988,
        <geoclide.shapes.DifferentialGeometry at 0x7f589349cf40>,
        True)
        >>> thit, dg, is_int = sph1.intersect(r)
        >>> dg.p # the intersection point
        Point(-0.6000000000000121, 0.0, 0.8)
        >>> dg.n # The surface normal at the intersection point
        Normal(-0.6, 0.0, 0.8000000000000002)
        >>> sph2.intersect(r) # here no intersection since the sphere part above z=0.5 is removed
        (None, None, False)
        """
        if not isinstance(r1, Ray): raise ValueError('The given parameter must be a Ray')
        ray = Ray(r1)
        ray.o = self.wTo[r1.o]
        ray.d = self.wTo[r1.d]

        # Compute quadratic sphere coefficients
        a = ray.d.x*ray.d.x + ray.d.y*ray.d.y + ray.d.z*ray.d.z
        b = 2 * (ray.d.x*ray.o.x + ray.d.y*ray.o.y + ray.d.z*ray.o.z)
        c = ray.o.x*ray.o.x + ray.o.y*ray.o.y + ray.o.z*ray.o.z - \
            self.radius*self.radius

        # Solve quadratic equation
        exist, t0, t1 = quadratic(a, b, c)
        if (not exist): return None, None, False
        
        # Compute intersection distance along ray
        if (t0 > ray.maxt or t1 < ray.mint): return None, None, False
        thit = t0

        if (t0 < ray.mint):
            thit = t1
            if (thit > ray.maxt): return None, None, False

        # Compute sphere hit position and $\phi$
        phit = ray[thit]
        if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
        phi = math.atan2(phit.y, phit.x)
        if (phi < 0): phi += 2*math.pi

        # Test sphere intersection against clipping parameters
        if ((self.zmin > -self.radius and phit.z < self.zmin) or
            (self.zmax <  self.radius and phit.z > self.zmax) or
            (phi > self.phiMax) ):
            if (thit == t1): return None, None, False
            if (t1 > ray.maxt): return None, None, False
            thit = t1
            # Compute sphere hit position and $\phi$
            phit = ray[thit]
            if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
            phi = math.atan2(phit.y, phit.x)
            if (phi < 0): phi += 2*math.pi
            if ((self.zmin > -self.radius and phit.z < self.zmin) or
                (self.zmax <  self.radius and phit.z > self.zmax) or
                (phi > self.phiMax) ):
                return None, None, False

        # Find parametric representation of sphere hit
        u = phi / self.phiMax
        theta = math.acos(clamp(phit.z / self.radius, -1, 1))
        v = (theta - self.thetaMin) / (self.thetaMax - self.thetaMin)

        # Compute sphere $\dpdu$ and $\dpdv$
        zradius = math.sqrt(phit.x*phit.x + phit.y*phit.y)
        invzradius = 1 / zradius
        cosphi = phit.x * invzradius
        sinphi = phit.y * invzradius
        dpdu = Vector(-self.phiMax * phit.y, self.phiMax * phit.x, 0)
        dpdv = Vector(Vector(phit.z * cosphi, phit.z * sinphi, -self.radius \
                        * math.sin(theta)) * (self.thetaMax-self.thetaMin))

        # Initialize _DifferentialGeometry_ from parametric information
        dg = DifferentialGeometry(self.oTw[phit], self.oTw[dpdu], self.oTw[dpdv], u, v, self)

        return thit, dg, True

    def area(self):
        return (self.phiMax * self.radius * (self.zmax-self.zmin)) # The sphere / partial sphere area


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.basic import Vector, Point, Normal
import math
import numpy as np


def dot(a, b):
    """
    The dot/scalar product

    Definition of the dot product:
    - (a . b) = a.x*b.x + a.y*b.y + a.z*b.z
    - (a . b) = ||a|| * ||b|| * cos(θ)

    Parameters
    ----------
    a : Vector | Normal
        The first vector or normal used for the dot product
    b : Vector | Normal
        The second vector or normal used for the dot product
    
    Results
    -------
    out : float
        The result of the dot product i.e. sum of products

    Examples
    --------
    >>> import geoclide as gc
    >>> a = gc.Vector(0., 0., 1.)
    >>> b = gc.Vector(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    >>> gc.dot(a,b)
    0.7071067811865476
    """
    if (isinstance(a, Vector) or isinstance(a, Normal)) and \
       (isinstance(b, Vector) or isinstance(b, Normal)):
        return (a.x*b.x + a.y*b.y + a.z*b.z)
    else:
        raise ValueError('Only Vector or Normal parameters are accepted')


def cross(a, b):
    """
    The cross product

    - /!\\ The cross product of 2 normals is not allowed

    Definition of the cross product:
    - (a × b) = ((a.y*b.z)-(a.z*b.y))x̂ + ((a.z*b.x)-(a.x*b.z))ŷ + ((a.x*b.y)-(a.y*b.x))*ẑ
    - (a × b) = ||a|| * ||b|| * sin(θ)

        where x̂, ŷ and ẑ are the unitary vectors respectively in axes x, y and z

    Parameters
    ----------
    a : Vector | Normal
        The first vector or normal used for the cross product
    b : Vector | Normal
        The second vector or normal used for the cross product
    
    Results
    -------
    out : Vector
        The result of the cross product

    Examples
    --------
    >>> import geoclide as gc
    >>> a = gc.Vector(0.,0.,1.)
    >>> b = gc.Vector(1.,0.,0.)
    >>> gc.cross(a,b)
    Vector(0.0, 1.0, 0.0)
    """
    if ( (isinstance(a, Vector) and isinstance(b, Vector)) or 
         (isinstance(a, Vector) and isinstance(b, Normal)) or 
         (isinstance(a, Normal) and isinstance(b, Vector)) ):
        return Vector((a.y*b.z)-(a.z*b.y), (a.z*b.x)-(a.x*b.z), (a.x*b.y)-(a.y*b.x))
    elif isinstance(a, Normal) and isinstance(b, Normal):
        raise ValueError('Only 1 Normal is tolerated not 2')
    else:
        raise ValueError('Only Vector or Normal parameters are accepted')
    

def normalize(v):
    """
    Normalize a Vector/Normal

    Parameters
    ----------
    v : Vector | Normal
        The Vector or Normal to be normalized
    
    Results
    -------
    out : Vector | Normal
        The normalized vector/normal

    Examples
    --------
    >>> import geoclide as gc
    >>> v = gc.Vector(1.,0.,1.)
    >>> gc.normalize(v)
    Vector(0.7071067811865475, 0.0, 0.7071067811865475)
    """
    if isinstance(v, Vector) or isinstance(v, Normal):
        return v / v.length()
    else:
        raise ValueError('The parameter v must be a Vector or Normal')


def coordinate_system(v1, method="m2"):
    """
    Create an orthogonal coordinate system from 1 vector (v1)

    Parameters
    ----------
    v1 : Vector
        The base vector used to create the orthogonal coordinate system
    method: str, optional
        Default is 'm2' (method from pbrt v4), other choice is 'm1' (pbrt v2 and v3)
    
    Results
    -------
    v2 : Vector
        The second vector of the orthogonal coordinate system
    v3 : Vector
        The third vector of the orthogonal coordinate system
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(0., 0., 1.)
    >>> v2, v3 = gc.coordinate_system(v1)
    (Vector(1.0, -0.0, -0.0), Vector(-0.0, 1.0, -0.0))
    """
    if not isinstance(v1, Vector):
        raise ValueError("The parameter v1 must be a Vector")

    # used in pbrt v4    
    if (method == "m2"):
        if v1.z > 0:sign = 1.
        else: sign =-1.
        term_1 = -1. / (sign + v1.z)
        term_2 = v1.x * v1.y * term_1
        v2 = Vector(1 + sign * (v1.x*v1.x) * term_1, sign * term_2, -sign * v1.x)
        v3 = Vector(term_2, sign + (v1.y*v1.y) * term_1, -v1.y)
    # used in pbrt v2 and v3
    elif (method == "m1"):
        if (abs(v1.x) > abs(v1.y)):
            invLen = 1/ math.sqrt(v1.x*v1.x + v1.z*v1.z)
            v2 = Vector(-v1.z*invLen, 0, v1.x*invLen)
        else:
            invLen = 1/ math.sqrt(v1.y*v1.y + v1.z*v1.z)
            v2 = Vector(0, v1.z*invLen, -v1.y*invLen)
        v3 = cross(v1, v2)
    else:
        raise ValueError("Only 2 choices for parameter method: 'm1' or 'm2'")
    
    return v2, v3


def distance(p1, p2):
    """
    Compute the distance between 2 points

    Parameters
    ----------
    p1 : Point
        The first point
    p2 : Point
        The second point

    Results
    -------
    out : float
        The distance between the 2 points
    
    Examples
    --------
    >>> p1 = gc.Point(0., 0., 0.)
    >>> p2 = gc.Point(0., 0., 10.)
    >>> gc.distance(p1,p2)
    10.0
    >>> p1 = gc.Point(1., 2., 1.9)
    >>> p2 = gc.Point(5., 15., 3.)
    >>> gc.distance(p1,p2)
    13.64587849865299
    """
    if isinstance(p1, Point) and isinstance(p2, Point):
        return (p1 - p2).length()
    else:
        raise ValueError('Only Point parameters are accepted')


def face_forward(a, b):
    """
    Flip the Vector/Normal a if the Vector/Normal b is in the opposite direction

    It can be useful to flip a surface normal so that it lies in the same
    hemisphere as a given vector.

    Parameters
    ----------
    a : Vector | Normal
        The Vector or Normal to potentially flip
    b : Vector | Normal
        The base Vector or Normal used for the flip

    Results
    -------
    out : Vector | Normal
        The potentially flipped Vector or Normal

    Examples
    --------
    >>> import geoclide as gc
    >>> n1 = gc.Normal(1., 0., 0.)
    >>> v1 = gc.Vector(-1., 0., 0.)
    >>> gc.face_forward(v1, n1)
    Vector(1.0, -0.0, -0.0)
    """
    if (isinstance(a, Vector) or isinstance(a, Normal)) and \
    (isinstance(b, Vector) or isinstance(b, Normal)):
        return (a*-1) if (dot(a, b) < 0) else a
    else:
        raise ValueError('Only Vector or Normal parameters are accepted')


def vmax(a):
    """
    Returns the largest component value of the Vector/Point/Normal

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector/point/normal used
    
    Results
    -------
    out: float
        The largest vector/point/normal value
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2.,3.,1.)
    >>> gc.max(v1)
    3
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        return int(np.max(a.to_numpy()))
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')


def vmin(a):
    """
    Returns the smallest component value of the Vector/Point/Normal

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector/point/normal used
    
    Results
    -------
    out: float
        The smallest vector/point/normal value
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2.,3.,1.)
    >>> gc.min(v1)
    1
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        return int(np.min(a.to_numpy()))
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')
    

def vargmax(a):
    """
    Returns the index of the Vector/Point/Normal component with the largest value

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector/point/normal used
    
    Results
    -------
    out: int
        The index of the largest vector/point/normal value
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2.,3.,1.)
    >>> gc.argmax(v1)
    1
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        return (0 if a.x>a.z else 2) if (a.x>a.y) else (1 if a.y>a.z else 2)
        #return int(np.argmax(a.to_numpy()))
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')


def vargmin(a):
    """
    Returns the index of the Vector/Point/Normal component with the smallest value

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector/point/normal used
    
    Results
    -------
    out: int
        The index of the smallest vector/point/normal value
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2.,3.,1.)
    >>> gc.argmin(v1)
    2
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        return (0 if a.x<a.z else 2) if (a.x<a.y) else (1 if a.y<a.z else 2)
        #return int(np.argmin(a.to_numpy()))
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')
    

def vabs(a):
    """
    Calculate the absolute value of each components of the vector/point/normal

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector/point/normal used
    
    Results
    -------
    out: in progress...
    
    """
    if isinstance(a, Vector):
        return Vector(abs(a.x), abs(a.y), abs(a.z))
    elif isinstance(a, Point):
        return Point(abs(a.x), abs(a.y), abs(a.z))
    elif isinstance(a, Normal):
        return Normal(abs(a.x), abs(a.y), abs(a.z))
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')


def permute(a, ix, iy=None, iz=None):
    """
    Permutes the vector/point/normal values according to the given indices

    Parameters
    ----------
    a : Vector | Point | Normal
    ix : int | np.ndarray
        The index of the value we want to keep as a remplacement for the x component.
        If ix is an np.ndarray -> the 1st component is ix and circumvent iy and iz parameters by the 2nd and 3th array components
    iy : int, optional
        The index of the value we want to keep as a remplacement for the y component.
    iz : int, optional
        The index of the value we want to keep as a remplacement for the z component.
    
    Results
    -------
    out : Vector | Point | Normal
        The vector/point/normal after the permute operation

    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2., 3., 1.)
    >>> gc.permute(v1, 1, 0, 2)
    Vector(3.0, 2.0, 1.0)
    >>> gc.permute(v1, np.array([1, 0, 2]))
    Vector(3.0, 2.0, 1.0)
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        if (isinstance(ix, np.ndarray)):
            if len(ix) != 3:
                raise ValueError("The size of the given array must be 3")
            if (  not ( (isinstance(ix[0], np.integer)) and
                       (isinstance(ix[1], np.integer)) and
                       (isinstance(ix[2], np.integer)) )  ):
                raise ValueError("The components of the given array must be integers")
            return Vector(a[ix[0]], a[ix[1]], a[ix[2]])
        elif( (isinstance(ix, int) or isinstance(ix, np.integer)) and
              (isinstance(iy, int) or isinstance(iy, np.integer)) and
              (isinstance(iz, int) or isinstance(iz, np.integer)) ):
            return Vector(a[ix], a[iy], a[iz])
        else:
            raise ValueError("Wrong parameter value(s) for ix and/or iy and/or iz")
    else:
        raise NameError('The parameter a must be a Vector or Point or Normal')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.basic import Ray, BBox
from geoclide.quadrics import Sphere
from geoclide.trianglemesh import Triangle, TriangleMesh
import xarray as xr
import numpy as np
from datetime import datetime
from geoclide.constante import VERSION

def calc_intersection(shape, r1, method='v3'):
    """
    Performs intersection test between a shape and a ray and returns dataset

    Parameters
    ----------
    shape : BBox | Sphere | Triangle | TriangleMesh
        The shape used for the intersection
    r1 : Ray
        The ray used for the iuntersection
    method : str, optional
        Used in triangle intersection test. Only two choice : 'v2' and 'v3'.
        The 'v3' have more robustness tests, and 'v2' is faster.

    Returns
    ------
    out : xr.Dataset
        Look-up table with the intersection information
    
    Examples
    --------
    >>> import geoclide as gc
    >>> sphere = gc.Sphere(radius=1.) # sphere of radius 1
    >>> bbox = gc.BBox(p1=gc.Point(0., 0., 0.), p2=gc.Point(1.,1.,1.))
    >>> ray = gc.Ray(o=gc.Point(-2., 0., 0.8), d=gc.Vector(1.,0.,0.))
    >>> ds_sphere = gc.calc_intersection(sphere, ray)
    >>> ds_sphere
    <xarray.Dataset>
    Dimensions:          (xyz: 3, dim_0: 4, dim_1: 4)
    Coordinates:
    * xyz              (xyz) int64 0 1 2
    Dimensions without coordinates: dim_0, dim_1
    Data variables: (12/17)
        is_intersection  bool True
        o                (xyz) float64 -2.0 0.0 0.8
        d                (xyz) float64 1.0 0.0 0.0
        mint             int64 0
        maxt             float64 inf
        shape            <U6 'Sphere'
        ...               ...
        oTw_mInv         (dim_0, dim_1) float64 1.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 1.0
        wTo_m            (dim_0, dim_1) float64 1.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 1.0
        wTo_mInv         (dim_0, dim_1) float64 1.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 1.0
        thit             float64 1.4
        phit             (xyz) float64 -0.6 0.0 0.8
        nhit             (xyz) float64 -0.6 0.0 0.8
    >>> ds_box = gc.calc_intersection(bbox, ray)
    >>> ds_bbox
    <xarray.Dataset>
    Dimensions:          (xyz: 3)
    Coordinates:
    * xyz              (xyz) int64 0 1 2
    Data variables:
        is_intersection  bool True
        o                (xyz) float64 -2.0 0.0 0.8
        d                (xyz) float64 1.0 0.0 0.0
        mint             int64 0
        maxt             float64 inf
        shape            <U4 'BBox'
        pmin             (xyz) float64 0.0 0.0 0.0
        pmax             (xyz) float64 1.0 1.0 1.0
        thit             float64 2.0
        phit             (xyz) float64 0.0 0.0 0.8
    """
    if (not isinstance(r1, Ray)):
        raise ValueError('The parameter r1 must a Ray')

    if (isinstance(shape, BBox)):
        t0, t1, is_intersection = shape.intersect(r1)
        if is_intersection:
            if t0 > 0: thit = t0
            else: thit = t1
            phit = r1[thit]
            nhit = None # TODO compute the real normal
        else:
            thit = None
            phit = None
            nhit = None
    elif(isinstance(shape, Sphere)      or
         isinstance(shape, Triangle)    or 
         isinstance(shape, TriangleMesh)):
        if (isinstance(shape, Triangle) or isinstance(shape, TriangleMesh)):
            thit, dg, is_intersection = shape.intersect(r1, method=method)
        else:
            thit, dg, is_intersection = shape.intersect(r1)
        if is_intersection:
            phit = dg.p
            nhit = dg.n
        else:
            phit = None
            nhit = None
    else:
        raise ValueError('The only supported shape are: BBox and Sphere')
    
    ds = xr.Dataset(coords={'xyz':np.arange(3)})
    ds['is_intersection'] = is_intersection
    ds['is_intersection'].attrs = {'description':'tells if there is an intersection between the ray and the shape'}

    ds['o'] = xr.DataArray(r1.o.to_numpy(), dims='xyz')
    ds['o'].attrs = {'type': 'Point', 'description':'the x, y and z components of the ray point'}
    ds['d'] = xr.DataArray(r1.d.to_numpy(), dims='xyz')
    ds['d'].attrs = {'type': 'Vector', 'description':'the x, y and z components of the ray vector'}
    ds['mint'] = r1.mint
    ds['mint'].attrs = {'description':'the mint attribut of the ray'}
    ds['maxt'] = r1.maxt
    ds['maxt'].attrs = {'description':'the maxt attribut of the ray'}

    if (isinstance(shape, BBox)):
        ds.attrs = {'shape': 'BBox'}
        ds['pmin'] = xr.DataArray(shape.pmin.to_numpy(), dims='xyz')
        ds['pmin'].attrs = {'type': 'Point', 'description':'the x, y and z components of the pmin BBox attribut'}
        ds['pmax'] = xr.DataArray(shape.pmax.to_numpy(), dims='xyz')
        ds['pmax'].attrs = {'type': 'Point', 'description':'the x, y and z components of the pmax BBox attribut'}
    if (isinstance(shape, Sphere)):
        ds.attrs = {'shape':  'Sphere'}
        ds['radius'] = shape.radius
        ds['radius'].attrs = {'description':'the sphere radius attribut'}
        ds['z_min'] = shape.zmin
        ds['z_min'].attrs = {'description':'the sphere zmin attribut'}
        ds['z_max'] = shape.zmax
        ds['z_max'].attrs = {'description':'the sphere zmax attribut'}
        ds['phi_max'] = shape.phiMax
        ds['phi_max'].attrs = {'unit':'Radian', 'description':'the sphere phiMax attribut'}
    if (isinstance(shape, Triangle)):
        ds.attrs = {'shape': 'Triangle'}
        ds['p0'] = xr.DataArray(shape.p0.to_numpy(), dims='xyz')
        ds['p0'].attrs = {'description': 'the triangle p0 attribut'}
        ds['p1'] = xr.DataArray(shape.p1.to_numpy(), dims='xyz')
        ds['p1'].attrs = {'description': 'the triangle p1 attribut'}
        ds['p2'] = xr.DataArray(shape.p2.to_numpy(), dims='xyz')
        ds['p2'].attrs = {'description': 'the triangle p2 attribut'}
    if (isinstance(shape, TriangleMesh)):
        ds.attrs = {'shape': 'TriangleMesh'}
        ds['v'] = xr.DataArray(np.array([pi.to_numpy() for pi in shape.vertices]), dims=['nvertices', 'xyz'])
        ds['v'].attrs = {'description': 'The vertices xyz coordinates.'}
        ds['vi'] = xr.DataArray(shape.vertices_index.reshape(shape.ntriangles, 3), dims=['ntriangles', 'p0p1p2'])
        ds['vi'].attrs = {'description': 'For each triangle, the index of vertices point p0, p1 and p2 (from variable v).'}
        ds.attrs.update({'ntriangles': shape.ntriangles,
                         'nvertices' : shape.nvertices})
    if (not isinstance(shape, BBox)):
        ds['wTo_m'] = xr.DataArray(shape.wTo.m)
        ds['wTo_m'].attrs = {'description':'the transformation matrix of the ' + str(ds.attrs['shape']).lower() + ' wTo attribut'}
        ds['wTo_mInv'] = xr.DataArray(shape.wTo.mInv)
        ds['wTo_mInv'].attrs = {'description':'the inverse transformation matrix of the ' + str(ds.attrs['shape']).lower() + ' wTo attribut'}
        ds['oTw_m'] = xr.DataArray(shape.oTw.m)
        ds['oTw_m'].attrs = {'description':'the transformation matrix of the ' + str(ds.attrs['shape']).lower() + ' oTw attribut'}
        ds['oTw_mInv'] = xr.DataArray(shape.oTw.mInv)
        ds['oTw_mInv'].attrs = {'description':'the inverse transformation matrix of the ' + str(ds.attrs['shape']).lower() +' oTw attribut'}


    if (thit is not None):
        ds['thit'] = thit
        ds['thit'].attrs = {'description':'the t ray factor for the intersection point calculation'}
    if (phit is not None):
        ds['phit'] = xr.DataArray(phit.to_numpy(), dims='xyz')
        ds['phit'].attrs = {'type': 'Point', 'description':'the x, y and z components of the intersection point'}
    if (nhit is not None):
        ds['nhit'] = xr.DataArray(nhit.to_numpy(), dims='xyz')
        ds['nhit'].attrs = {'type': 'Normal', 'description':'the x, y and z components of the normal at the intersection point'}

    date = datetime.now().strftime("%Y-%m-%d")  
    ds.attrs.update({'date':date,
                     'version': VERSION})

    return ds


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.mathope import gamma_f32, gamma_f64


VERSION = '1.0.0'

GAMMA2_F32 = gamma_f32(2)
GAMMA2_F64 = gamma_f64(2)

GAMMA3_F32 = gamma_f32(3)
GAMMA3_F64 = gamma_f64(3)

GAMMA5_F32 = gamma_f32(5)
GAMMA5_F64 = gamma_f64(5)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import math
import os
ROOTPATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
import sys
sys.path.insert(0, ROOTPATH)
import geoclide as gc

P1 = [np.array([0.,0.,0.]), np.array([1.,2.,3.])]
V1 = [np.array([0.,0.,1.]), np.array([1.,1.,1.])]
N1 = [np.array([0.,0.,1.]), np.array([1.,1.,1.])]


@pytest.mark.parametrize('v_arr', V1)
def test_vector(v_arr):
    v1 = gc.Vector(v_arr[0], v_arr[1], v_arr[2])
    v2 = gc.Vector(v_arr)
    assert (v1 == v2)
    assert (np.all(v1.to_numpy() == v_arr))
    assert (v1[0] == v_arr[0])
    assert (v1[1] == v_arr[1])
    assert (v1[2] == v_arr[2])
    assert (v1[0] == v1.x)
    assert (v1[1] == v1.y)
    assert (v1[2] == v1.z)
    assert (-v1 == gc.Vector(-v1.x, -v1.y, -v1.z))


@pytest.mark.parametrize('p_arr', P1)
def test_point(p_arr):
    p1 = gc.Point(p_arr[0], p_arr[1], p_arr[2])
    p2 = gc.Point(p_arr)
    assert (p1 == p2)
    assert (np.all(p1.to_numpy() == p_arr))
    assert (p1[0] == p_arr[0])
    assert (p1[1] == p_arr[1])
    assert (p1[2] == p_arr[2])
    assert (p1[0] == p1.x)
    assert (p1[1] == p1.y)
    assert (p1[2] == p1.z)
    assert (-p1 == gc.Point(-p1.x, -p1.y, -p1.z))


@pytest.mark.parametrize('n_arr', N1)
def test_normal(n_arr):
    n1 = gc.Normal(n_arr[0], n_arr[1], n_arr[2])
    n2 = gc.Normal(n_arr)
    assert (n1 == n2)
    assert (np.all(n1.to_numpy() == n_arr))
    assert (n1[0] == n_arr[0])
    assert (n1[1] == n_arr[1])
    assert (n1[2] == n_arr[2])
    assert (n1[0] == n1.x)
    assert (n1[1] == n1.y)
    assert (n1[2] == n1.z)
    assert (-n1 == gc.Normal(-n1.x, -n1.y, -n1.z))


@pytest.mark.parametrize('v_arr', V1)
def test_ope_vector(v_arr):
    v1 = gc.Vector(v_arr)
    v2 = gc.Vector(1., 3., 0.5)
    assert (v1+v2 == gc.Vector(v_arr+v2.to_numpy()))
    assert (v1-v2 == gc.Vector(v_arr-v2.to_numpy()))
    assert (v1*2 == gc.Vector(v_arr*2.))
    assert (2*v1 == gc.Vector(v_arr*2.))
    assert (v1/2 == gc.Vector(v_arr/2.))
    v3 = v1+v2
    assert (v3.length() == math.sqrt(v3[0]**2+v3[1]**2+v3[2]**2))
    assert (v3.length_squared() == v3[0]**2+v3[1]**2+v3[2]**2)


@pytest.mark.parametrize('p_arr', P1)
def test_ope_point(p_arr):
    p1 = gc.Point(p_arr)
    v1 = gc.Vector(1., 3., 0.5)
    p2 = gc.Point(1.,1.,1.)
    assert (p1+v1 == gc.Point(p_arr+v1.to_numpy()))
    assert (p1-p2 == gc.Vector(p_arr-p2.to_numpy()))
    assert (p1-v1 == gc.Point(p_arr-v1.to_numpy()))
    assert (p1*2 == gc.Point(p_arr*2.))
    assert (2*p1 == gc.Point(p_arr*2.))
    assert (p1/2 == gc.Point(p_arr/2.))


@pytest.mark.parametrize('n_arr', N1)
def test_ope_normal(n_arr):
    n1 = gc.Normal(n_arr)
    n2 = gc.Normal(1., 3., 0.5)
    assert (n1+n2 == gc.Normal(n_arr+n2.to_numpy()))
    assert (n1-n2 == gc.Normal(n_arr-n2.to_numpy()))
    assert (n1*2 == gc.Normal(n_arr*2.))
    assert (2*n1 == gc.Normal(n_arr*2.))
    assert (n1/2 == gc.Normal(n_arr/2.))
    assert (n1.length() == math.sqrt(n1[0]**2+n1[1]**2+n1[2]**2))
    assert (n1.length_squared() == n1[0]**2+n1[1]**2+n1[2]**2)


@pytest.mark.parametrize('p_arr', P1)
@pytest.mark.parametrize('v_arr', V1)
def test_ray(p_arr, v_arr):
    p1 = gc.Point(p_arr)
    v1 = gc.Vector(v_arr)
    r1 = gc.Ray(p1,v1)
    assert (r1.o == p1)
    assert (r1.d == v1)
    assert (r1[10.] == gc.Point(p_arr + 10.*v_arr))
    assert (r1.mint == 0)
    assert (r1.maxt == float("inf"))
    r2 = gc.Ray(p1,v1,0.5, 20.)
    assert (r2.mint == 0.5)
    assert (r2.maxt == 20.)
    r3 = gc.Ray(r2)
    assert (r3.o == r2.o)
    assert (r3.d == r2.d)
    assert (r3.mint == r2.mint)
    assert (r3.maxt == r2.maxt)


def test_bbox():
    # test attributs
    p1 = gc.Point(1., 1., 1.)
    p2 = gc.Point(2., 2., 3.)
    b1 = gc.BBox(p1, p2)
    minx = min(p1.x, p2.x)
    miny = min(p1.y, p2.y)
    minz = min(p1.z, p2.z)
    maxx = max(p1.x, p2.x)
    maxy = max(p1.y, p2.y)
    maxz = max(p1.z, p2.z)
    pmin = gc.Point(minx, miny, minz)
    pmax = gc.Point(maxx, maxy, maxz)
    assert (b1.pmin == pmin)
    assert (b1.pmax == pmax)
    p = [pmin]
    p.append(gc.Point(pmax.x,pmin.y,pmin.z))
    p.append(gc.Point(pmax.x,pmax.y,pmin.z))
    p.append(gc.Point(pmin.x,pmax.y,pmin.z))
    p.append(gc.Point(pmin.x,pmin.y,pmax.z))
    p.append(gc.Point(pmax.x,pmin.y,pmax.z))
    p.append(pmax)
    p.append(gc.Point(pmin.x,pmax.y,pmax.z))
    assert (b1.p0 == p[0])
    assert (b1.p1 == p[1])
    assert (b1.p2 == p[2])
    assert (b1.p3 == p[3])
    assert (b1.p4 == p[4])
    assert (b1.p5 == p[5])
    assert (b1.p6 == p[6])
    assert (b1.p7 == p[7])
    for i in range (len(b1.vertices)):
        assert (b1.vertices[i] == p[i])

    # test union method
    p3 = gc.Point(3., 3., 3.)
    b2 = b1.union(p3)
    assert (b2.pmin.x == min(b1.pmin.x, p3.x))
    assert (b2.pmin.y == min(b1.pmin.y, p3.y))
    assert (b2.pmin.z == min(b1.pmin.z, p3.z))
    assert (b2.pmax.x == max(b1.pmax.x, p3.x))
    assert (b2.pmax.y == max(b1.pmax.y, p3.y))
    assert (b2.pmax.z == max(b1.pmax.z, p3.z))
    b3 = b1.union(b2)
    assert (b3.pmin.x == min(b1.pmin.x, b2.pmin.x))
    assert (b3.pmin.y == min(b1.pmin.y, b2.pmin.y))
    assert (b3.pmin.z == min(b1.pmin.z, b2.pmin.z))
    assert (b3.pmax.x == max(b1.pmax.x, b2.pmax.x))
    assert (b3.pmax.y == max(b1.pmax.y, b2.pmax.y))
    assert (b3.pmax.z == max(b1.pmax.z, b2.pmax.z))

    # test is_inside method
    pIn = gc.Point(1.5, 1.5, 1.5)
    pOut = gc.Point(0., 0., 0.)
    assert (b3.is_inside(pIn))
    assert (not b3.is_inside(pOut))

    # test common_vertices method annd function get_common_vertices
    bc1 = gc.BBox(gc.Point(0., 0., 0.), gc.Point(2.5, 2.5, 2.5))
    bc2 = gc.BBox(gc.Point(2.5, 0., 0.), gc.Point(5., 2.5, 2.5))
    # The vertices of bc1 in common with the bc2 vertices are p1, p2, p5 and p6: 
    assert (np.all(bc1.common_vertices(bc2) == np.array([False, True, True, False, False, True, True, False])))
    assert (np.all(gc.get_common_vertices(bc1, bc2) == np.array([False, True, True, False, False, True, True, False])))
    # The vertices of bc2 in common with the bc1 vertices are p0, p3, p4 and p7: 
    assert (np.all(bc2.common_vertices(bc1) == np.array([True, False, False, True, True, False, False, True])))
    assert (np.all(gc.get_common_vertices(bc2, bc1) == np.array([True, False, False, True, True, False, False, True])))

    # test common_face method
    b_ref = gc.BBox(gc.Point(0., 0., 0.), gc.Point(1., 1., 1.))
    b_f0 = gc.BBox(gc.Point(1., 0., 0.), gc.Point(2., 1., 1.))
    b_f1 = gc.BBox(gc.Point(-1., 0., 0.), gc.Point(0., 1., 1.))
    b_f2 = gc.BBox(gc.Point(0., 1., 0.), gc.Point(1., 2., 1.))
    b_f3 = gc.BBox(gc.Point(0., -1., 0.), gc.Point(1., 0., 1.))
    b_f4 = gc.BBox(gc.Point(0., 0., 1.), gc.Point(1., 1., 2.))
    b_f5 = gc.BBox(gc.Point(0., 0., -1.), gc.Point(1., 1., 0.))
    assert (b_ref.common_face(b_f0) == 0)
    assert (b_ref.common_face(b_f1) == 1)
    assert (b_ref.common_face(b_f2) == 2)
    assert (b_ref.common_face(b_f3) == 3)
    assert (b_ref.common_face(b_f4) == 4)
    assert (b_ref.common_face(b_f5) == 5)
    assert (gc.get_common_face(b_ref, b_f0) == 0)
    assert (gc.get_common_face(b_ref, b_f1) == 1)
    assert (gc.get_common_face(b_ref, b_f2) == 2)
    assert (gc.get_common_face(b_ref, b_f3) == 3)
    assert (gc.get_common_face(b_ref, b_f4) == 4)
    assert (gc.get_common_face(b_ref, b_f5) == 5)

    # test method intersect_p
    b_int = gc.BBox(gc.Point(0., 0., 0.), gc.Point(1., 1., 1.))
    r_int_1 = gc.Ray(gc.Point(-0.5, 0.5, 0.5), gc.Vector(1.,0.,0.))
    r_int_2 = gc.Ray(gc.Point(0.5, 0.5, 0.5), gc.Vector(1.,0.,1.))
    r_int_3 = gc.Ray(gc.Point(0.5, 0.5, 2.), gc.Vector(0.,0.,1.))
    # case where the ray origin is outside and where the ray intersect 2 times the BBox
    t0, t1, is_int = b_int.intersect(r_int_1)
    p_int_t0 = r_int_1[t0]
    p_int_t1 = r_int_1[t1]
    assert (is_int)
    assert (np.isclose(p_int_t0.x, 0., 0., 1e-14))
    assert (np.isclose(p_int_t0.y, 0.5, 0., 1e-14))
    assert (np.isclose(p_int_t0.z, 0.5, 0., 1e-14))
    assert (np.isclose(p_int_t1.x, 1., 0., 1e-14))
    assert (np.isclose(p_int_t1.y, 0.5, 0., 1e-14))
    assert (np.isclose(p_int_t1.z, 0.5, 0., 1e-14))
    # case where the ray origin is inside and where the ray intersect 1 times the BBox
    t0, t1, is_int = b_int.intersect(r_int_2)
    p_int_t1 = r_int_2[t1]
    assert (is_int)
    assert (np.isclose(p_int_t1.x, 1., 0., 1e-14))
    assert (np.isclose(p_int_t1.y, 0.5, 0., 1e-14))
    assert (np.isclose(p_int_t1.z, 1.0, 0., 1e-14))
    # case where the ray origin is outsie and where the ray does not intersect with the BBox
    t0, t1, is_int = b_int.intersect(r_int_3)
    assert (not is_int)

    b_int2 = gc.BBox(p1=gc.Point(-500., -500., 0.), p2=gc.Point(500., 500., 700.))
    r_int4 = gc.Ray(o=gc.Point(0., 0., 0.), d=gc.normalize(gc.Vector(0.5, -0.5, 1.)))

    t0, t1, is_int = b_int2.intersect(r_int4)
    assert (is_int)
    assert (t0 == 0.)
    assert (np.isclose(857.3214099741128, t1, 0., 1e-14))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import math
import os
ROOTPATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
import sys
sys.path.insert(0, ROOTPATH)
import geoclide as gc

V1 = [np.array([5.,0.,0.]), np.array([5.,3.,1.])]
V2 = [np.array([5.,1.,1.]), np.array([5.,3.,2.])]
ANGLES = [30., 45., 60.]


@pytest.mark.parametrize('v_arr', V1)
def test_get_translate_tf(v_arr):
    t = gc.get_translate_tf(gc.Vector(v_arr))
    m = np.identity(4)
    m[0,-1] = v_arr[0]
    m[1,-1] = v_arr[1]
    m[2,-1] = v_arr[2]
    assert (np.all(t.m == m))
    mInv = np.identity(4)
    mInv[0,-1] = -v_arr[0]
    mInv[1,-1] = -v_arr[1]
    mInv[2,-1] = -v_arr[2]
    assert (np.all(t.mInv == mInv))


@pytest.mark.parametrize('v_arr', V2)
def test_get_scale_tf(v_arr):
    t = gc.get_scale_tf(v_arr[0], v_arr[1], v_arr[2])
    m = np.identity(4)
    m[0,0] = v_arr[0]
    m[1,1] = v_arr[1]
    m[2,2] = v_arr[2]
    assert (np.all(t.m == m))
    mInv = np.identity(4)
    mInv[0,0] = 1 * (1/v_arr[0])
    mInv[1,1] = 1 * (1/v_arr[1])
    mInv[2,2] = 1 * (1/v_arr[2])
    assert (np.all(t.mInv == mInv))


@pytest.mark.parametrize('angle', ANGLES)
def test_get_rotateX_tf(angle):
    t = gc.get_rotateX_tf(angle)
    sin_t = math.sin(angle*(math.pi / 180.))
    cos_t = math.cos(angle*(math.pi / 180.))
    m = np.identity(4)
    m[1,1] = cos_t
    m[1,2] = -1.*sin_t
    m[2,1] = sin_t
    m[2,2] = cos_t
    assert (np.all(t.m == m))
    mInv = np.transpose(m)
    assert (np.all(t.mInv == mInv))


@pytest.mark.parametrize('angle', ANGLES)
def test_get_rotateY_tf(angle):
    t = gc.get_rotateY_tf(angle)
    sin_t = math.sin(angle*(math.pi / 180.))
    cos_t = math.cos(angle*(math.pi / 180.))
    m = np.identity(4)
    m[0,0] = cos_t
    m[2,0] = -1.*sin_t
    m[0,2] = sin_t
    m[2,2] = cos_t
    assert (np.all(t.m == m))
    mInv = np.transpose(m)
    assert (np.all(t.mInv == mInv))


@pytest.mark.parametrize('angle', ANGLES)
def test_get_rotateZ_tf(angle):
    t = gc.get_rotateZ_tf(angle)
    sin_t = math.sin(angle*(math.pi / 180.))
    cos_t = math.cos(angle*(math.pi / 180.))
    m = np.identity(4)
    m[0,0] = cos_t
    m[0,1] = -1.*sin_t
    m[1,0] = sin_t
    m[1,1] = cos_t
    assert (np.all(t.m == m))
    mInv = np.transpose(m)
    assert (np.all(t.mInv == mInv))


@pytest.mark.parametrize('angle', ANGLES)
def test_get_rotate_tf_1(angle):
    v_x = gc.Vector(1., 0., 0.)
    tx = gc.get_rotateX_tf(angle)
    t = gc.get_rotate_tf(angle, v_x)
    assert (np.all(tx.m == t.m))
    assert (np.all(tx.mInv == t.mInv))

    v_y = gc.Vector(0., 1., 0.)
    ty = gc.get_rotateY_tf(angle)
    t = gc.get_rotate_tf(angle, v_y)
    assert (np.all(ty.m == t.m))
    assert (np.all(ty.mInv == t.mInv))

    v_z = gc.Vector(0., 0., 1.)
    tz = gc.get_rotateZ_tf(angle)
    t = gc.get_rotate_tf(angle, v_z)
    assert (np.all(tz.m == t.m))
    assert (np.all(tz.mInv == t.mInv))


@pytest.mark.parametrize('angle', ANGLES)
@pytest.mark.parametrize('v_arr', V1)
def test_get_rotate_tf_2(angle, v_arr):
    v = gc.normalize(gc.Vector(v_arr)) # important to normalize for Rodrigues formula
    t = gc.get_rotate_tf(angle, v)
    # Use independant method -> Rodrigues rotation formula
    m = np.identity(4)
    cos_t = math.cos(angle*(math.pi / 180.))
    sin_t = math.sqrt(1.-cos_t*cos_t)
    matA = np.identity(3)
    matB = np.zeros((3,3))
    matB[0,1] = -v.z
    matB[0,2] =  v.y
    matB[1,0] =  v.z
    matB[1,2] = -v.x
    matB[2,0] = -v.y
    matB[2,1] =  v.x
    matC = matB.dot(matB)
    m[0:3,0:3] = (matA + matB*sin_t + matC*(1-cos_t))
    assert (np.allclose(t.m, m, rtol=0., atol=1e-15))
    mInv = np.transpose(m)
    assert (np.allclose(t.mInv, mInv, rtol=0., atol=1e-15))


def test_transform():
    t1 = gc.Transform()
    assert (np.all(t1.m == np.identity(4)))
    assert (np.all(t1.mInv == np.identity(4)))

    t2 = gc.get_translate_tf(gc.Vector(5., 5., 5.))
    p1 = gc.Point(0., 0., 0.)
    v1 = gc.normalize(gc.Vector(1., 1., 1.))
    n1 = gc.Normal(0.,0.,1.)
    b1 = gc.BBox(p1,p1+v1)
    b1_bis = gc.BBox()
    b1_bis = b1_bis.union(t2[b1.p0])
    b1_bis = b1_bis.union(t2[b1.p1])
    b1_bis = b1_bis.union(t2[b1.p2])
    b1_bis = b1_bis.union(t2[b1.p3])
    b1_bis = b1_bis.union(t2[b1.p4])
    b1_bis = b1_bis.union(t2[b1.p5])
    b1_bis = b1_bis.union(t2[b1.p6])
    b1_bis = b1_bis.union(t2[b1.p7])
    assert (t2[p1] == gc.Point(5., 5., 5.))
    assert (t2[v1] == v1)
    assert (t2[n1] == n1)
    assert (t2[b1].pmin == b1_bis.pmin)
    assert (t2[b1].pmax == b1_bis.pmax)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide import Vector, Point, Normal
from geoclide import normalize, cross
import numpy as np
from geoclide.transform import Transform


class Shape(object):
    '''
    The parent class of the different shape classes
    '''
    indShape = 0
    def __init__(self, ObjectToWorld, WorldToObject):
        if (not isinstance(ObjectToWorld, Transform) or not isinstance(WorldToObject, Transform)):
            raise ValueError('The parameters oTw and wTo must be both Transform')
        self.oTw = ObjectToWorld
        self.wTo = WorldToObject


class DifferentialGeometry(object):
    '''
    The general parametric description of most of the shapes

    - A point p is described by a function depending to variables u and v such that p=f(u,v)
    - From u and v we can get two directions (partial derivative of p) parallel to the shape surface

    Parameters
    ----------
    p : Point, optional
        The concerned position, at the surface of the given shape
    dpdu : Vector, optional
        The surface partial derivative of p with respect to u
    dpdv : vector, optional
        The surface partial derivative of p with respect to v
    u : float, optional
        The u coordinate of the parametric representation
    v : float, optional
        The v coordinate of the parametric representation
    shape : Sphere | Triangle | ...
        The shape used
    '''
    def __init__(self, p=None, dpdu=None, dpdv=None, u=None, v=None, shape = None):
        if ( (p is None) and (dpdu is None) and (dpdv is None) and (u is None) and (v is None) and (shape is None) ):
            self.p = Point(0., 0., 0.)
            self.dpdu = Vector(0., 0., 0.)
            self.dpdv = Vector(0., 0., 0.)
            self.n = Normal(0., 0., 0.)
            self.u = 0.
            self.v = 0.
            self.shape = None
        elif ( isinstance(p, Point)     and
               isinstance(dpdu, Vector) and
               isinstance(dpdv, Vector) and
               np.isscalar(u)           and
               np.isscalar(v) ):
            self.p = p
            self.dpdu = dpdu
            self.dpdv = dpdv
            self.n = Normal(normalize(cross(self.dpdu, self.dpdv)))
            self.u = u
            self.v = v
            self.shape = shape
        else:
            raise ValueError('Problem with parameter(s)')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide import Vector, Point, Normal, Ray, BBox
from geoclide import normalize
import numpy as np
from numpy.linalg import inv
import math


class Transform(object):
    '''
    Tool to perform tranlation(s) and/or rotation(s) to objects

    Parameters
    ----------
    m : Transform | np.ndarray, optional
    mInv : np.ndarray, optional

    Exemples
    --------
    >>> import geoclide as gc
    >>> t1 = gc.Transform()
    >>> t1
    m=
    array(
    [[1. 0. 0. 0.]
    [0. 1. 0. 0.]
    [0. 0. 1. 0.]
    [0. 0. 0. 1.]] )
    mInv=
    array(
    [[1. 0. 0. 0.]
    [0. 1. 0. 0.]
    [0. 0. 1. 0.]
    [0. 0. 0. 1.]] )
    '''

    def __init__(self, m = None, mInv = None):
        if (isinstance(m, Transform)):
            self.m = m.m
            self.mInv = m.mInv
        elif (m is None and mInv is None):
            self.m = np.identity(4)
            self.mInv = self.m.copy()
        elif (isinstance(m, np.ndarray) and mInv is None):
            if (m.shape != (4,4)):
                raise ValueError("The m parameter must be an np.array of shape (4,4)")
            self.m = m
            self.mInv = inv(m)
        elif (m is None and isinstance(mInv, np.ndarray)):
            if (mInv.shape != (4,4)):
                raise ValueError("The mInv parameter must be an np.array of shape (4,4)")
            self.m = inv(m)
            self.mInv = mInv
        elif (isinstance(m, np.ndarray) and isinstance(mInv, np.ndarray)):
            if (m.shape != (4,4) or mInv.shape != (4,4)):
                raise ValueError("The matrix shape of m and mInv must be (4,4)")
            self.m = m
            self.mInv = mInv
        else:
            raise ValueError("Wrong parameter value(s) for Transform")

    def __eq__(self, t):
        if (not isinstance(t, Transform)):
            raise ValueError("Equality with a Transform must be only with another Transform")
        
        self.m = t.m
        self.mInv = t.mInv
        
    def __mul__(self, T): 
        if (not isinstance(T, Transform)):
            raise ValueError('A transform can be multiplied only by another Transform')
        
        return Transform(np.dot(self.m, T.m), np.dot(T.mInv, self.mInv))
    
    def __getitem__(self, c):
        """"
        Apply the transformations

        Parameters
        ----------
        c : Vector | Point | Normal | Ray | BBox
            The Vector/Point/Normal/Ray/BBox to which the transformation is applied
        
        Results
        -------
        out : Vector | Point | Normal | Ray | BBox
            The Vector/Point/Normal/Ray/BBox after the transformation

        Examples
        --------
        >>> import geoclide as gc
        >>> t = gc.get_translate_tf(gc.Vector(5., 5., 5.))
        >>> p = gc.Point(0., 0., 0.)
        >>> t[p]
        Point(5.0, 5.0, 5.0)
        """
        if isinstance(c, Vector):
            xv = self.m[0,0]*c.x + self.m[0,1]*c.y + self.m[0,2]*c.z
            yv = self.m[1,0]*c.x + self.m[1,1]*c.y + self.m[1,2]*c.z
            zv = self.m[2,0]*c.x + self.m[2,1]*c.y + self.m[2,2]*c.z
            return Vector(xv, yv, zv)
        elif isinstance(c, Point):
            xp = self.m[0,0]*c.x + self.m[0,1]*c.y + self.m[0,2]*c.z + self.m[0,3]
            yp = self.m[1,0]*c.x + self.m[1,1]*c.y + self.m[1,2]*c.z + self.m[1,3]
            zp = self.m[2,0]*c.x + self.m[2,1]*c.y + self.m[2,2]*c.z + self.m[2,3]
            wp = self.m[3,0]*c.x + self.m[3,1]*c.y + self.m[3,2]*c.z + self.m[3,3]
            if (wp == 1):
                return Point(xp, yp, zp)
            else: 
                return Point(xp, yp, zp)/wp
        elif isinstance(c, Normal):
            xn = self.mInv[0,0]*c.x + self.mInv[1,0]*c.y + self.mInv[2,0]*c.z
            yn = self.mInv[0,1]*c.x + self.mInv[1,1]*c.y + self.mInv[2,1]*c.z
            zn = self.mInv[0,2]*c.x + self.mInv[1,2]*c.y + self.mInv[2,2]*c.z
            return Normal(xn, yn, zn)
        elif isinstance(c, Ray):
            R = Ray(c.o, c.d)
            R.o = self[R.o]
            R.d = self[R.d]
            return R
        elif isinstance(c, BBox):
            b = BBox()
            p0 = self[c.p0]
            v0 = self[c.p1-c.p0]
            v1 = self[c.p3-c.p0]
            v2 = self[c.p4-c.p0]
            b = b.union(p0)
            b = b.union(p0+v0)
            b = b.union(p0+(v0+v1))
            b = b.union(p0+v1)
            b = b.union(p0+v2)
            b = b.union(p0+(v0+v2))
            b = b.union(p0+(v0+v1+v2))
            b = b.union(p0+(v1+v2))
            return b
        else:
            raise ValueError('Unknown type for transformations')

    def __str__(self):
        print("m=\n", self.m, "\nmInv=\n", self.mInv)
        return ""
    
    def __repr__(self):
        print("m=\narray(\n", self.m, ")\nmInv=\narray(\n",self.mInv,")")
        return ""

    def inverse(self):
        """
        Inverse the initial transformation

        Parameters
        ----------
        t : Transform
            The transformation to be inversed

        Returns
        -------
        out : Transform
            The inversed transformation
        """
        return get_inverse_tf(self)

    def isIdentity(self):
        return (self.m[0,0] == 1) and (self.m[0,1] == 0) and (self.m[0,2] == 0) and \
            (self.m[0,3] == 0) and (self.m[1,0] == 0) and (self.m[1,1] == 1) and \
            (self.m[1,2] == 0) and (self.m[1,3] == 0) and (self.m[2,0] == 0) and \
            (self.m[2,1] == 0) and (self.m[2,2] == 1) and (self.m[2,3] == 0) and \
            (self.m[3,0] == 0) and (self.m[3,1] == 0) and (self.m[3,2] == 0) and \
            (self.m[3,3] == 1)

    def translate(self, v):
        """
        Apply translate to initial transformation
        
        Parameters
        ----------
        v : Vector
            The vector used for the transformation

        Returns
        -------
        t : Transform
            The product of the initial transformation and the translate transformation

        examples
        --------
        >>> import geoclide as gc
        >>> t = Transform()
        >>> t = t.translate(gc.Vector(5.,0.,0.))
        >>> t
        m=
        array(
        [[1. 0. 0. 5.]
        [0. 1. 0. 0.]
        [0. 0. 1. 0.]
        [0. 0. 0. 1.]] )
        mInv=
        array(
        [[ 1.  0.  0. -5.]
        [ 0.  1.  0.  0.]
        [ 0.  0.  1.  0.]
        [ 0.  0.  0.  1.]] )
        """
        t = get_translate_tf(v)
        return self*t

    def scale(self, x, y, z):
        """
        Apply scale to initial transformation

        Parameters
        ----------
        x : float
            The scale factor to apply (x axis)
        y : float
            The scale factor to apply (y axis)
        y : float
            The scale factor to apply (z axis)

        Returns
        -------
        t : Transform
            The product of the initial transformation and the scale transformation
        """
        t = get_scale_tf(x,y,z)
        return self*t

    def rotateX(self, angle):
        """
        Apply rotateX to initial transformation

        Parameters
        ----------
        angle : float
            The angle in degrees for the rotation around the x axis

        Returns
        -------
        t : Transform
            The product of the initial transformation and the rotateX transformation
        """
        t = get_rotateX_tf(angle)
        return self*t

    def rotateY(self, angle):
        """
        Apply rotateY to initial transformation

        Parameters
        ----------
        angle : float
            The angle in degrees for the rotation around the y axis

        Returns
        -------
        t : Transform
            The product of the initial transformation and the rotateY transformation
        """
        t = get_rotateY_tf(angle)
        return self*t

    def rotateZ(self, angle):
        """
        Apply rotateZ to initial transformation

        Parameters
        ----------
        v : Vector
            The angle in degrees for the rotation around the Z axis

        Returns
        -------
        t : Transform
            The product of the initial transformation and the rotateZ transformation
        """
        t = get_rotateZ_tf(angle)
        return self*t

    def rotate(self, angle, axis):
        """
        Apply rotate to initial transformation

        Parameters
        ----------
        angle : float
            The angle in degrees for the rotation
        axis : Vector | Normal
            The rotation is performed arount the parameter axis 

        Returns
        -------
        t : Transform
            The product of the initial transformation and the rotate transformation
        """
        t = get_rotate_tf(angle, axis)
        return self*t


def get_inverse_tf(t):
    """
    Get the inverse transformation

    Parameters
    ----------
    t : Transform
        The transformation to be inversed

    Returns
    -------
    out : Transform
        The inversed transformation
    """
    return Transform(t.mInv, t.m)
    

def get_translate_tf(v):
    """
    Get the translate Transform

    Parameters
    ----------
    v : Vector
        The vector used for the transformation

    Returns
    -------
    t : Transform
        The translate transformation

    examples
    --------
    >>> import geoclide as gc
    >>> t = gc.get_translate_tf(gc.Vector(5.,0.,0.))
    >>> t
    m=
    array(
    [[1. 0. 0. 5.]
    [0. 1. 0. 0.]
    [0. 0. 1. 0.]
    [0. 0. 0. 1.]] )
    mInv=
    array(
    [[ 1.  0.  0. -5.]
    [ 0.  1.  0. -0.]
    [ 0.  0.  1. -0.]
    [ 0.  0.  0.  1.]] )
    """
    if (not isinstance(v, Vector)):
        raise ValueError("The parameter v must be a Vector")
    
    m = np.identity(4)
    m[0,3] = v.x
    m[1,3] = v.y
    m[2,3] = v.z
    mInv = np.identity(4)
    mInv[0,3] = (v.x)*-1
    mInv[1,3] = (v.y)*-1
    mInv[2,3] = (v.z)*-1
    return Transform(m, mInv)


def get_scale_tf(x, y, z):
    """
    Get the scale Transform

    Parameters
    ----------
    x : float
        The scale factor to apply (x axis)
    y : float
        The scale factor to apply (y axis)
    y : float
        The scale factor to apply (z axis)

    Returns
    -------
    t : Transform
        The scale transformation
    """
    if ( (not np.isscalar(x)) or
         (not np.isscalar(y)) or
         (not np.isscalar(z)) ):
        raise ValueError("The parameters x, y and z must be all scalars")
    
    m = np.identity(4)
    m[0,0] = x
    m[1,1] = y
    m[2,2] = z
    mInv = np.identity(4)
    mInv[0,0] = 1./x
    mInv[1,1] = 1./y
    mInv[2,2] = 1./z
    return Transform(m, mInv)


def get_rotateX_tf(angle):
    """
    Get the rotateX Transform

    Parameters
    ----------
    angle : float
        The angle in degrees for the rotation around the x axis

    Returns
    -------
    t : Transform
        The rotateX transformation
    """
    if (not np.isscalar(angle)):
        raise ValueError("The parameter angle must be a scalar")
    
    sin_t = math.sin(angle*(math.pi / 180.))
    cos_t = math.cos(angle*(math.pi / 180.))
    m = np.identity(4)
    m[1,1] = cos_t
    m[1,2] = -1.*sin_t
    m[2,1] = sin_t
    m[2,2] = cos_t
    return Transform(m, np.transpose(m))


def get_rotateY_tf(angle):
    """
    Get the rotateY Transform

    Parameters
    ----------
    angle : float
        The angle in degrees for the rotation around the y axis

    Returns
    -------
    t : Transform
        The rotateY transformation
    """
    if (not np.isscalar(angle)):
        raise ValueError("The parameter angle must be a scalar")
    
    sin_t = math.sin(angle*(math.pi / 180.))
    cos_t = math.cos(angle*(math.pi / 180.))
    m = np.identity(4)
    m[0,0] = cos_t
    m[2,0] = -1.*sin_t
    m[0,2] = sin_t
    m[2,2] = cos_t
    return Transform(m, np.transpose(m))


def get_rotateZ_tf(angle):
    """
    Get the rotateZ Transform

    Parameters
    ----------
    v : Vector
        The angle in degrees for the rotation around the Z axis

    Returns
    -------
    t : Transform
        The rotateZ transformation
    """
    if (not np.isscalar(angle)):
        raise ValueError("The parameter angle must be a scalar")
    
    sin_t = math.sin(angle*(math.pi / 180.))
    cos_t = math.cos(angle*(math.pi / 180.))
    m = np.identity(4)
    m[0,0] = cos_t
    m[0,1] = -1.*sin_t
    m[1,0] = sin_t
    m[1,1] = cos_t
    return Transform(m, np.transpose(m))


def get_rotate_tf(angle, axis):
    """
    Get the rotate Transform around a given axis

    Parameters
    ----------
    angle : float
        The angle in degrees for the rotation
    axis : Vector | Normal
        The rotation is performed around the Vector/Normal axis 

    Returns
    -------
    t : Transform
        The rotate transformation
    """
    if (not np.isscalar(angle)):
        raise ValueError("The parameter angle must be a scalar")
    if ( (not isinstance(axis, Vector)) and
         (not isinstance(axis, Normal)) ):
        raise ValueError("The parameter axis must be a Vector or a Normal")

    a = Vector(normalize(axis))
    s = math.sin(angle*(math.pi / 180.))
    c = math.cos(angle*(math.pi / 180.))
    m = np.identity(4)

    m[0,0] = a.x*a.x+(1-a.x*a.x)*c
    m[0,1] = a.x*a.y*(1-c)-a.z*s
    m[0,2] = a.x*a.z*(1-c)+a.y*s

    m[1,0] = a.x*a.y*(1-c)+a.z*s
    m[1,1] = a.y*a.y+(1-a.y*a.y)*c
    m[1,2] = a.y*a.z*(1-c)-a.x*s

    m[2,0] = a.x*a.z*(1-c)-a.y*s
    m[2,1] = a.y*a.z*(1-c)+a.x*s
    m[2,2] = a.z*a.z+(1-a.z*a.z)*c
    return Transform(m, np.transpose(m))





#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.shapes import Shape, DifferentialGeometry
from geoclide.basic import Vector, Point, Ray
import geoclide.vecope as gv
import numpy as np
from geoclide.constante import GAMMA2_F64, GAMMA3_F64, GAMMA5_F64
from geoclide.transform import Transform


class Triangle(Shape):
    '''
    Creation of the class Triangle

    Parameters
    ----------
    p0 : Point
        The first point of the triangle
    p1 : Point
        The second point of the triangle
    p2 : Point
        The the third point of the triangle
    oTw : Transform, optional
        From object to world space or the transformation applied to the triangle
    wTo : Transform, optional
        From world to object space or the in inverse transformation applied to the triangle
    p0t : Point, optional
        If given circumvent the automatically computed p0t (p0 after applying transformation)
    p1t : Point, optional
        If given circumvent the automatically computed p1t (p1 after applying transformation)
    p2t : Point, optional
        If given circumvent the automatically computed p2t (p2 after applying transformation)
    '''
    def __init__(self, p0=None, p1=None, p2=None, oTw=None, wTo=None,
                 p0t=None, p1t=None, p2t=None):
        # Manage None cases
        if p0 is None : p0 = Point()
        if p1 is None : p1 = Point()
        if p2 is None : p2 = Point()
        if oTw is None and wTo is None:
            oTw = Transform()
            wTo = Transform()
            self.p0t = p0
            self.p1t = p1
            self.p2t = p2
        elif ( (oTw is None or isinstance(oTw, Transform)) and
               (wTo is None or isinstance(wTo, Transform)) ):
            if (oTw is None): oTw = wTo.inverse() # if oTw is None then wTo should be Transform
            if (wTo is None): wTo = oTw.inverse() # if wTo is None then oTw should be Transform
            if (p0t is None): self.p0t = oTw[p0]
            if (p1t is None): self.p1t = oTw[p1]
            if (p2t is None): self.p2t = oTw[p2]

        if (not isinstance(p0, Point) or not isinstance(p1, Point) or not isinstance(p2, Point)):
            raise ValueError('The parameters p0, p1 and p2 must be all Point')
        if ( (p0t is not None and not isinstance(p0t, Point)) or
             (p1t is not None and not isinstance(p1t, Point)) or
             (p2t is not None and not isinstance(p2t, Point)) ):
            raise ValueError('The parameters p0t, p1t and p2t must be all Point')
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        self.p0 = p0
        self.p1 = p1
        self.p2 = p2
        if (p0t is not None): self.p0t = p0t
        if (p1t is not None): self.p1t = p1t
        if (p2t is not None): self.p2t = p2t

    def is_intersection(self, r1, method='v3'):
        """
        Test if a Ray intersect with the triangle

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test
        method : str, optional
            Tow choice -> 'v2' (use mainly pbrt v2 intersection test method) or 'v3' (pbrt v3)

        Returns
        -------
        out : bool
            If there is an intersection -> True, else False
        """
        if method == 'v3':
            return self.is_intersection_v3(r1)
        elif method == 'v2':
            return self.is_intersection_v2(r1)
        else:
            raise ValueError("Only 'v2' and 'v3' are valid values for method parameter")   
    
    def intersect(self, r1, method='v3'):
        """
        Test if a Ray intersect with the triangle and return intersection information

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test
        method : str, optional
            Tow choice -> 'v2' (use mainly pbrt v2 intersection test method) or 'v3' (pbrt v3)
        
        Returns
        -------
        thit : float
            The t ray variable for its first intersection at the shape surface
        dg : DifferentialGeometry
            The parametric parameters at the intersection point
        is_intersection : bool
            If there is an intersection -> True, else False

        Notes
        -----
        By default the 'v3' method is used since there are more robustness tests.
        But the 'v2' method is at least twice faster than 'v3'.
        """
        if method == 'v3':
            return self.intersect_v3(r1)
        elif method == 'v2':
            return self.intersect_v2(r1)
        else:
            raise ValueError("Only 'v2' and 'v3' are valid values for method parameter")   
        
    def intersect_v2(self, r1):
        """
        Test if a Ray intersect with the triangle using mainly pbrt v2 method,
        and return intersection information
        
        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test
        
        Returns
        -------
        thit : float
            The t ray variable for its first intersection at the shape surface
        dg : DifferentialGeometry
            The parametric parameters at the intersection point
        is_intersection : bool
            If there is an intersection -> True, else False
        """
        if not isinstance(r1, Ray): raise ValueError('The given parameter must be a Ray')
        ray = Ray(r1)
        p0 = self.p0t
        p1 = self.p1t
        p2 = self.p2t
        e1 = p1 - p0
        e2 = p2 - p0
        s1 = gv.cross(ray.d, e2)
        divisor = gv.dot(s1, e1)

        if (divisor == 0):
            return None, None, False
        invDivisor = 1./divisor

        # compute the first barycentric coordinate
        s = ray.o - p0
        b1 = gv.dot(s, s1) * invDivisor
        if (b1 < -0.00000001 or  b1 > 1.00000001):
            return None, None, False

        # compute the second barycentric coordinate
        s2 = gv.cross(s, e1)
        b2 = gv.dot(ray.d, s2) * invDivisor
        if (b2 < 0 or  b1+b2 > 1):
            return None, None, False

        # compute the time at the intersection point
        t = gv.dot(e2, s2) * invDivisor
        if (t < ray.mint or t > ray.maxt):
            return None, None, False

        # compute triangle partial derivatives
        uvs = np.array([[0., 0.], [1., 0.], [1., 1.]])

        # compute deltas for triangle partial derivatives
        du1 = uvs[0][0] - uvs[2][0]
        du2 = uvs[1][0] - uvs[2][0]
        dv1 = uvs[0][1] - uvs[2][1]
        dv2 = uvs[1][1] - uvs[2][1]
        dp1 = p0 - p2
        dp2 = p1 - p2
        determinant = du1 * dv2 - dv1 * du2

        if (determinant == 0):
            dpdu, dpdv = gv.coordinate_system(gv.normalize(gv.cross(e2, e1)))
        else:
            invdet = 1./determinant
            dpdu = ( dp1*dv2   - dp2*dv1) * invdet
            dpdv = (dp1*(-du2) + dp2*du1) * invdet
        
        # interpolate $(u,v)$ triangle parametric coordinates
        b0 = 1 - b1 - b2
        tu = b0*uvs[0][0] + b1*uvs[1][0] + b2*uvs[2][0]
        tv = b0*uvs[0][1] + b1*uvs[1][1] + b2*uvs[2][1]

        # fill the DifferentialGeometry and thit
        dg = DifferentialGeometry(ray[t], dpdu, dpdv, tu, tv, self)
        thit = t

        return thit, dg, True
    
    def intersect_v3(self, r1):
        """
        Test if a Ray intersect with the triangle using mainly pbrt v3 method,
        and return intersection information

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test
        
        Returns
        -------
        thit : float
            The t ray variable for its first intersection at the shape surface
        dg : DifferentialGeometry
            The parametric parameters at the intersection point
        is_intersection : bool
            If there is an intersection -> True, else False
        """
        if not isinstance(r1, Ray): raise ValueError('The given parameter must be a Ray')
        ray = Ray(r1)
        p0 = self.p0t
        p1 = self.p1t
        p2 = self.p2t

        # Get triangle vertices and translate them in based on ray origin
        p0t = p0 - ray.o
        p1t = p1 - ray.o
        p2t = p2 - ray.o

        kz = gv.vargmax(gv.vabs(ray.d))
        kx = kz + 1
        if(kx == 3): kx = 0
        ky = kx + 1
        if(ky == 3): ky = 0

        d = gv.permute(ray.d, kx, ky, kz)
        p0t = gv.permute(p0t, kx, ky, kz)
        p1t = gv.permute(p1t, kx, ky, kz)
        p2t = gv.permute(p2t, kx, ky, kz)
        
        sx = -d.x/d.z
        sy = -d.y/d.z
        sz = 1./d.z
        p0t.x += sx*p0t.z
        p0t.y += sy*p0t.z
        p1t.x += sx*p1t.z
        p1t.y += sy*p1t.z
        p2t.x += sx*p2t.z
        p2t.y += sy*p2t.z
        
        # Compute edge function coefficients
        e0 = (p1t.x * p2t.y) - (p1t.y * p2t.x)
        e1 = (p2t.x * p0t.y) - (p2t.y * p0t.x)
        e2 = (p0t.x * p1t.y) - (p0t.y * p1t.x)

        # Perform triangle edge and determinant tests
        if ((e0 < 0 or e1 < 0 or e2 < 0) and (e0 > 0 or e1 > 0 or e2 > 0)):
            return None, None, False
        det = e0 + e1 + e2
        if (det == 0): return None, None, False

        # Compute scaled hit distance to triangle and test against ray $t$ range
        p0t.z *=  sz
        p1t.z *=  sz
        p2t.z *=  sz

        tScaled = e0*p0t.z + e1*p1t.z + e2*p2t.z

        if ( (det < 0 and (tScaled >= 0 or tScaled < ray.maxt*det)) or
             (det > 0 and (tScaled <= 0 or tScaled > ray.maxt*det)) ):
            return None, None, False

        # Compute barycentric coordinates and t value for triangle intersection
        invDet = 1./det
        b0 = e0 * invDet
        b1 = e1 * invDet
        b2 = e2 * invDet
        t = tScaled * invDet
        
        # Ensure that computed triangle t is conservatively greater than zero
        maxZt = np.max(np.abs(np.array([p0t.z, p1t.z, p2t.z])))
        deltaZ = GAMMA3_F64 * maxZt
        maxXt = np.max(np.abs(np.array([p0t.x, p1t.x, p2t.x])))
        maxYt = np.max(np.abs(np.array([p0t.y, p1t.y, p2t.y])))
        deltaX = GAMMA5_F64 * (maxXt + maxZt)
        deltaY = GAMMA5_F64 * (maxYt + maxZt) 
        deltaE = 2 * (GAMMA2_F64 * maxXt * maxYt + deltaY * maxXt + deltaX * maxYt)
        maxE = np.max(np.abs(np.array([e0, e1, e2])))
        deltaT = 3 * (GAMMA3_F64 * maxE * maxZt + deltaE * maxZt + deltaZ * maxE) * abs(invDet)
        if (t <= deltaT): return None, None, False

        # Compute triangle partial derivatives
        # Below the z components is not needed since we are in 2D with u in x and v un y
        dpdu = Vector()
        dpdv = Vector()
        uv0 = Point(0., 0., 0.)
        uv1 = Point(1., 0., 0.)
        uv2 = Point(1., 1., 0.)
        duv02 = uv0 - uv2
        duv12 = uv1 - uv2
        dp02 = p0 - p2
        dp12 = p1 - p2
        determinant = duv02.x*duv12.y - duv02.y*duv12.x
        degenerate = bool(abs(determinant) < 1e-8)

        if (not degenerate):
            invdet = 1./ determinant
            dpdu = (duv12.y*dp02 - duv02.y*dp12)*invdet
            dpdv = (-duv12.x*dp02 + duv02.x*dp12)*invdet

        if ( degenerate or gv.cross(dpdu, dpdv).length_squared() == 0):
            ng = gv.cross(p2-p0, p1-p0)
            if ( ng.length_squared() == 0 ):
                return None, None, False
            dpdu, dpdv = gv.coordinate_system(gv.normalize(ng))

        phit = b0*p0+b1*p1+b2*p2
        uvhit =b0*uv0 + b1*uv1 + b2*uv2
        thit = t
        dg = DifferentialGeometry(phit, dpdu, dpdv, uvhit.x, uvhit.y, self)
        
        return thit, dg, True
    
    def is_intersection_v2(self, r1):
        """
        Test if a Ray intersect with the triangle using mainly pbrt v2 method

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test
        
        Returns
        -------
        out : bool
            If there is an intersection -> True, else False
        """
        if not isinstance(r1, Ray): raise ValueError('The given parameter must be a Ray')
        ray = Ray(r1)
        p0 = self.p0t
        p1 = self.p1t
        p2 = self.p2t
        e1 = p1 - p0
        e2 = p2 - p0
        s1 = gv.cross(ray.d, e2)
        divisor = gv.dot(s1, e1)

        if (divisor == 0):
            return False
        invDivisor = 1./divisor

        # compute the first barycentric coordinate
        s = ray.o - p0
        b1 = gv.dot(s, s1) * invDivisor
        if (b1 < -0.00000001 or  b1 > 1.00000001):
            return False

        # compute the second barycentric coordinate
        s2 = gv.cross(s, e1)
        b2 = gv.dot(ray.d, s2) * invDivisor
        if (b2 < 0 or  b1+b2 > 1):
            return False

        # compute the time at the intersection point
        t = gv.dot(e2, s2) * invDivisor
        if (t < ray.mint or t > ray.maxt):
            return False

        return True
    
    def is_intersection_v3(self, r1):
        """
        Test if a Ray intersect with the triangle using mainly pbrt v3 method

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test
        
        Returns
        -------
        out : bool
            If there is an intersection -> True, else False
        """
        if not isinstance(r1, Ray): raise ValueError('The given parameter must be a Ray')
        ray = Ray(r1)
        p0 = self.p0t
        p1 = self.p1t
        p2 = self.p2t

        # Get triangle vertices and translate them in based on ray origin
        p0t = p0 - ray.o
        p1t = p1 - ray.o
        p2t = p2 - ray.o

        kz = gv.vargmax(gv.vabs(ray.d))
        kx = kz + 1
        if(kx == 3): kx = 0
        ky = kx + 1
        if(ky == 3): ky = 0

        d = gv.permute(ray.d, kx, ky, kz)
        p0t = gv.permute(p0t, kx, ky, kz)
        p1t = gv.permute(p1t, kx, ky, kz)
        p2t = gv.permute(p2t, kx, ky, kz)
        
        sx = -d.x/d.z
        sy = -d.y/d.z
        sz = 1./d.z
        p0t.x += sx*p0t.z
        p0t.y += sy*p0t.z
        p1t.x += sx*p1t.z
        p1t.y += sy*p1t.z
        p2t.x += sx*p2t.z
        p2t.y += sy*p2t.z
        
        # Compute edge function coefficients
        e0 = (p1t.x * p2t.y) - (p1t.y * p2t.x)
        e1 = (p2t.x * p0t.y) - (p2t.y * p0t.x)
        e2 = (p0t.x * p1t.y) - (p0t.y * p1t.x)

        # Perform triangle edge and determinant tests
        if ((e0 < 0 or e1 < 0 or e2 < 0) and (e0 > 0 or e1 > 0 or e2 > 0)):
            return False
        det = e0 + e1 + e2
        if (det == 0): return False

        # Compute scaled hit distance to triangle and test against ray $t$ range
        p0t.z *=  sz
        p1t.z *=  sz
        p2t.z *=  sz

        tScaled = e0*p0t.z + e1*p1t.z + e2*p2t.z

        if ( (det < 0 and (tScaled >= 0 or tScaled < ray.maxt*det)) or
             (det > 0 and (tScaled <= 0 or tScaled > ray.maxt*det)) ):
            return False

        # Compute barycentric coordinates and t value for triangle intersection
        invDet = 1./det
        t = tScaled * invDet
        
        # Ensure that computed triangle t is conservatively greater than zero
        maxZt = np.max(np.abs(np.array([p0t.z, p1t.z, p2t.z])))
        deltaZ = GAMMA3_F64 * maxZt
        maxXt = np.max(np.abs(np.array([p0t.x, p1t.x, p2t.x])))
        maxYt = np.max(np.abs(np.array([p0t.y, p1t.y, p2t.y])))
        deltaX = GAMMA5_F64 * (maxXt + maxZt)
        deltaY = GAMMA5_F64 * (maxYt + maxZt)
        deltaE = 2 * (GAMMA2_F64 * maxXt * maxYt + deltaY * maxXt + deltaX * maxYt)
        maxE = np.max(np.abs(np.array([e0, e1, e2])))
        deltaT = 3 * (GAMMA3_F64 * maxE * maxZt + deltaE * maxZt + deltaZ * maxE) * abs(invDet)
        if (t <= deltaT): return False

        # Compute triangle partial derivatives
        # Below the z components is not needed since we are in 2D with u in x and v un y
        dpdu = Vector()
        dpdv = Vector()
        uv0 = Point(0., 0., 0.)
        uv1 = Point(1., 0., 0.)
        uv2 = Point(1., 1., 0.)
        duv02 = uv0 - uv2
        duv12 = uv1 - uv2
        dp02 = p0 - p2
        dp12 = p1 - p2
        determinant = duv02.x*duv12.y - duv02.y*duv12.x
        degenerate = bool(abs(determinant) < 1e-8)

        if (not degenerate):
            invdet = 1./ determinant
            dpdu = (duv12.y*dp02 - duv02.y*dp12)*invdet
            dpdv = (-duv12.x*dp02 + duv02.x*dp12)*invdet

        if ( degenerate or gv.cross(dpdu, dpdv).length_squared() == 0):
            ng = gv.cross(p2-p0, p1-p0)
            if ( ng.length_squared() == 0 ):
                return False
        
        return True
     

class TriangleMesh(Shape):
    '''
    Creation of the class TriangleMesh

    Parameters
    ----------
    vi : np.ndarray
        The 1d ndarray of size (3*ntriangles) containing the vertices indices of triangles (see the parameter v). 
        The 3 first indices are the vertices (p0, p1 and p3) indices of the first triangle and so on. 
        It can be a 2d ndarray of shape (ntriangles, 3)
    v : np.ndarray
        The vertices xyz coordinates. It is a 1d array of size (nvertices) containing Point objects 
        where the first element is the coordinate of first vertex and so on. It can be a 2d float array 
        of size (nvertices, 3).
    oTw : Transform, optional
        From object to world space or the transformation applied to the triangle mesh
    wTo : Transform, optional
        From world to object space or the in inverse transformation applied to the triangle mesh
    '''
    def __init__(self, vi, v, oTw=None, wTo=None):
        if (  not isinstance(vi, np.ndarray)                                or
              not (len(vi.shape) == 1 or len(vi.shape) == 2)                or
              not (np.issubdtype(vi.dtype, int) or np.issubdtype(vi.dtype, np.integer))  ):
            raise ValueError('The parameter vi must be a 1d or 2d ndarray of intergers')
        if (  not ( isinstance(v, np.ndarray) )                                              or
              not ( (len(v.shape) == 1 and isinstance(v[0], Point)) or (len(v.shape) == 2) )  ):
            raise ValueError('The paramerter v must be a 1d ndarray of Point objects or a 2d ndarray')
        
        if (len(vi) == 2): vi = vi.flatten()
        self.vertices_index = vi
        self.nvertices = np.amax(vi) + int(1)
        if not isinstance(v[0], Point):
            v_bis = np.empty((self.nvertices), dtype=Point)
            for iv in range (0, self.nvertices):
                v_bis[iv] = Point(v[iv,:])
            self.vertices = v_bis
        else:
            self.vertices = v
        
        if oTw is None:
            self.vertices_t = self.vertices
        else:
            self.vertices_t = np.empty((self.nvertices), dtype=Point)
            for iv in range (0, self.nvertices):
                self.vertices_t[iv] = oTw[self.vertices[iv]]
        
        self.ntriangles = int(len(np.atleast_1d(self.vertices_index))/3)
        self.triangles = np.empty((self.ntriangles), dtype=Triangle)
            
        for itri in range(0, self.ntriangles):
            p0 = self.vertices[self.vertices_index[int(3*itri)]]
            p1 = self.vertices[self.vertices_index[int(3*itri) + int(1)]]
            p2 = self.vertices[self.vertices_index[int(3*itri) + int(2)]]
            p0t = self.vertices_t[self.vertices_index[int(3*itri)]]
            p1t = self.vertices_t[self.vertices_index[int(3*itri) + int(1)]]
            p2t = self.vertices_t[self.vertices_index[int(3*itri) + int(2)]]
            self.triangles[itri] = Triangle(p0, p1, p2, oTw, wTo, p0t, p1t, p2t)
        
        Shape.__init__(self, ObjectToWorld = self.triangles[0].oTw,
                       WorldToObject = self.triangles[0].wTo)
    
    def intersect(self, r1, method='v3'):
        """
        Test if a Ray intersect with the triangle mesh and return intersection information

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test
        method : str, optional
            Tow choice -> 'v2' (use mainly pbrt v2 triangle intersection test method) or 'v3' (pbrt v3)
        
        Returns
        -------
        thit : float
            The t ray variable for its first intersection at the shape surface
        dg : DifferentialGeometry
            The parametric parameters at the intersection point
        is_intersection : bool
            If there is an intersection -> True, else False
        """
        dg = None
        thit = float("inf")
        for itri in range(0, self.ntriangles):
            thit_bis, dg_bis, is_intersection_bis = self.triangles[itri].intersect(r1, method=method)
            if is_intersection_bis:
                if thit > thit_bis:
                    thit = thit_bis
                    dg = dg_bis
        if dg is None: return None, None, False
            
        return thit, dg, True
    
    def is_intersection(self, r1, method='v3'):
        """
        Test if a Ray intersect with the triangle mesh

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test
        method : str, optional
            Tow choice -> 'v2' (use mainly pbrt v2 triangle intersection test method) or 'v3' (pbrt v3)
        
        Returns
        -------
        out : bool
            If there is an intersection -> True, else False
        """
        for itri in range(0, self.ntriangles):
            if (self.triangles[itri].is_intersection(r1, method=method)):
                return True
        return False


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math 
import numpy as np
from geoclide.constante import GAMMA3_F64


class Vector(object):
    """
    Parameters
    ----------
    x : float | Point | Vector | Normal | np.ndarray, optional
        If scalar -> x component of the vector.
        Else, circumvent the y and z parameters and take the components of the Point/Vector/Normal/np.ndarray.
    y : float, optional
        The y component of the vector.
    z : float, optional
        The z component of the vector.

    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(0.,0.,1.)
    >>> v1
    >>> Vector(0,0,1)
    """
    def __init__(self, x = 0., y = 0., z = 0.):
        if ( np.isscalar(x) and
             np.isscalar(y) and
             np.isscalar(z) ):
            self.x = float(x)
            self.y = float(y)
            self.z = float(z)
        elif ( isinstance(x, Vector) or
               isinstance(x, Point) or
               isinstance(x, Normal) ):
            self.x = x.x
            self.y = x.y
            self.z = x.z
        elif ( isinstance(x, np.ndarray) and
               len(x) == 3 ):
            self.x = float(x[0])
            self.y = float(x[1])
            self.z = float(x[2])
        else: 
            raise ValueError("Wrong parameter value(s)")

    def __eq__(self, v2):
        if isinstance(v2, Vector):
            return (self.x==v2.x) and (self.y==v2.y) and (self.z==v2.z)
        else:
            raise ValueError('Equality with a Vector must be only with another Vector')

    def __add__(self, v2):
        if isinstance(v2, Vector):
            return Vector(self.x+v2.x, self.y+v2.y, self.z+v2.z) 
        else:
            raise ValueError('Addition with a Vector must be only with another Vector')

    def __sub__(self, v2):
        if isinstance(v2, Vector):
            return Vector(self.x-v2.x, self.y-v2.y, self.z-v2.z)
        else:
            raise ValueError('Substraction with a Vector must be only with another Vector')

    def __truediv__(self, sca):
        if (np.isscalar(sca)):
            div = (1./sca)
            return Vector(self.x*div, self.y*div, self.z*div) 
        else:
            raise ValueError('A Vector can be divided only by a scalar')
    def __mul__(self, sca): 
        if (np.isscalar(sca)):
            return Vector(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise ValueError('A Vector can be multiplied only by a scalar')
        
    def __rmul__(self, sca): 
        if (np.isscalar(sca)):
            return Vector(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise ValueError('A Vector can be multiplied only by a scalar')
    
    def __neg__(self):
        return Vector(-self.x, -self.y, -self.z)

    def __getitem__(self, ind):
        if ( not isinstance(ind, int) or
             not isinstance(ind, np.integer) ):
            IndexError("Only an integer is a valid index")
        if ind == 0:
            return self.x
        elif ind == 1:
            return self.y
        elif ind == 2 :
            return self.z
        else:
            IndexError(f"Index {ind} is out of range") 

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'
    
    def __repr__(self):
        return 'Vector(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'
    
    def length_squared(self):
        return self.x*self.x + self.y*self.y + self.z*self.z
    
    def length(self):
        return math.sqrt(self.length_squared()) # L2 norm

    def to_numpy(self):
        return np.array([self.x, self.y, self.z], dtype=np.float64)
    

class Point(object):
    """
    Parameters
    ----------
    x : float | Point | Vector | Normal | np.ndarray, optional
        If scalar -> x component of the point.
        Else, circumvent the y and z parameters and take the components of the Point/Vector/Normal/np.ndarray.
    y : float, optional
        The y component of the point.
    z : float, optional
        The z component of the point.

    Examples
    --------
    >>> import geoclide as gc
    >>> p1 = gc.Point(0.,0.,1.)
    >>> p1
    >>> Point(0,0,1)
    """
    def __init__(self, x = 0., y = 0., z = 0.):
        if ( np.isscalar(x) and
             np.isscalar(y) and
             np.isscalar(z) ):
            self.x = float(x)
            self.y = float(y)
            self.z = float(z)
        elif ( isinstance(x, Vector) or
               isinstance(x, Point) or
               isinstance(x, Normal) ):
            self.x = x.x
            self.y = x.y
            self.z = x.z
        elif ( isinstance(x, np.ndarray) and
               len(x) == 3 ):
            self.x = float(x[0])
            self.y = float(x[1])
            self.z = float(x[2])
        else: 
            raise ValueError("Wrong parameter value(s)")

    def __eq__(self, p2):
        if isinstance(p2, Point):
            return (self.x==p2.x) and (self.y==p2.y) and (self.z==p2.z)
        else:
            raise ValueError('Equality with a Point must be only with another Point')

    def __add__(self, v):
        if isinstance(v, Vector) or isinstance(v, Point):
            return Point(self.x+v.x, self.y+v.y, self.z+v.z)
        else:
            raise ValueError('Addition with a Point must be only with a Vector or' +
                             ' (exceptionally tolerated) another Point')

    def __sub__(self, vp2):
        if isinstance(vp2, Vector):
            return Point(self.x-vp2.x, self.y-vp2.y, self.z-vp2.z)
        elif isinstance(vp2, Point):
            return Vector(self.x-vp2.x, self.y-vp2.y, self.z-vp2.z)
        else:
            raise ValueError('Substraction with a Point must be with another Point or a Vector')

    def __truediv__(self, sca):
        if (np.isscalar(sca)):
            div = (1./sca)
            return Point(self.x*div, self.y*div, self.z*div) 
        else:
            raise ValueError('A Point can be divided only by a scalar')

    def __mul__(self, sca): 
        if (np.isscalar(sca)):
            return Point(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise ValueError('A Point can be multiplied only by a scalar')
    
    def __rmul__(self, sca): 
        if (np.isscalar(sca)):
            return Point(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise ValueError('A Point can be multiplied only by a scalar')
    
    def __neg__(self):
        return Point(-self.x, -self.y, -self.z)
    
    def __getitem__(self, ind):
        if ( not isinstance(ind, int) or
             not isinstance(ind, np.integer) ):
            IndexError("Only an integer is a valid index")
        if ind == 0:
            return self.x
        elif ind == 1:
            return self.y
        elif ind == 2 :
            return self.z
        else:
            IndexError(f"Index {ind} is out of range")

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'
    
    def __repr__(self):
        return 'Point(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'

    def to_numpy(self):
        return np.array([self.x, self.y, self.z], dtype=np.float64)

class Normal(object):
    """
    Parameters
    ----------
    x : float | Point | Vector | Normal | np.ndarray, optional
        If scalar -> x component of the normal.
        Else, circumvent the y and z parameters and take the components of the Point/Vector/Normal/np.ndarray.
    y : float, optional
        The y component of the normal.
    z : float, optional
        The z component of the normal.

    Examples
    --------
    >>> import geoclide as gc
    >>> n1 = gc.Normal(0.,0.,1.)
    >>> n1
    >>> Normal(0,0,1)
    """
    def __init__(self, x = 0., y = 0., z = 0.):
        if ( np.isscalar(x) and
             np.isscalar(y) and
             np.isscalar(z) ):
            self.x = float(x)
            self.y = float(y)
            self.z = float(z)
        elif ( isinstance(x, Vector) or
               isinstance(x, Point) or
               isinstance(x, Normal) ):
            self.x = x.x
            self.y = x.y
            self.z = x.z
        elif ( isinstance(x, np.ndarray) and
               len(x) == 3 ):
            self.x = float(x[0])
            self.y = float(x[1])
            self.z = float(x[2])
        else: 
            raise ValueError("Wrong parameter value(s)")

    def __eq__(self, n2):
        if isinstance(n2, Normal):
            return (self.x==n2.x) and (self.y==n2.y) and (self.z==n2.z)
        else:
            raise ValueError('Equality with a Normal must be only with another Normal')

    def __add__(self, n2):
        if isinstance(n2, Normal):
            return Normal(self.x+n2.x, self.y+n2.y, self.z+n2.z) 
        else:
            raise ValueError('Addition with a Normal must be only with another Normal')

    def __sub__(self, n2):
        if isinstance(n2, Normal):
            return Normal(self.x-n2.x, self.y-n2.y, self.z-n2.z)
        else:
            raise ValueError('Substraction with a Normal must be only with another Normal')

    def __truediv__(self, sca):
        if (np.isscalar(sca)):
            div = (1./sca)
            return Normal(self.x*div, self.y*div, self.z*div) 
        else:
            raise ValueError('A Normal can be divided only by a scalar')

    def __mul__(self, sca):
        if (np.isscalar(sca)):
            return Normal(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise ValueError('A Normal can be multiplied only by a scalar')
    
    def __rmul__(self, sca):
        if (np.isscalar(sca)):
            return Normal(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise ValueError('A Normal can be multiplied only by a scalar')
    
    def __neg__(self):
        return Normal(-self.x, -self.y, -self.z)
        
    def __getitem__(self, ind):
        if ( not isinstance(ind, int) or
             not isinstance(ind, np.integer) ):
            IndexError("Only an integer is a valid index")
        if ind == 0:
            return self.x
        elif ind == 1:
            return self.y
        elif ind == 2 :
            return self.z
        else:
            IndexError(f"Index {ind} is out of range")

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'
    
    def __repr__(self):
        return 'Normal(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'
    
    def length_squared(self):
        return self.x*self.x + self.y*self.y + self.z*self.z
    
    def length(self):
        return math.sqrt(self.length_squared()) # L2 norm

    def to_numpy(self):
        return np.array([self.x, self.y, self.z], dtype=np.float64)


class Ray(object):
    """
    Definition of ray:

    r(t) = o + t*d, where:
    - o is the origin point of the ray
    - d is the direction of the ray
    - t is a positive real scalar

    Parameters
    ----------
    o : Point | Ray
        Origin point of the ray.
        If the o parameter is a Ray -> circumvent all the parameters by the ray attributs
    d : Vector
        Direction of the ray
    mint : float, optional
        The minimum t value
    maxt : float, optional
        The maximum t value

    Examples
    --------
    >>> import geoclide as gc
    >>> o = gc.Point(0., 50., 2.)
    >>> d = gc.Vector(0.,0.,1.)
    >>> r = gc.Ray(o, d, mint=20, maxt=100)
    >>> r
    r(t) = (0.0, 50.0, 2.0) + t*(0.0, 0.0, 1.0) with t ∈ [20,100[
    """
    def __init__(self, o, d=None, mint = 0, maxt = float("inf")):
        if isinstance(o, Ray):
            self.o = o.o
            self.d = o.d
            self.mint = float(o.mint)
            self.maxt = float(o.maxt)
        else:
            if (not isinstance(o, Point)):
                raise ValueError("The parameter o must be a Point or a Ray")
            if (not isinstance(d, Vector)):
                raise ValueError("The parameter d must only be a Vector")
            if (not np.isscalar(mint) or not np.isscalar(maxt)):
                raise ValueError("The parameters mint and maxt must be both scalars")
            if (mint > maxt):
                raise ValueError("maxt must be greater than mint")
            self.o = o
            self.d = d
            self.mint = mint
            self.maxt = maxt

    def __getitem__(self, t):
        if (not np.isscalar(t)):
            raise ValueError('The value must be a scalar')
        elif ( t < self.mint or t > self.maxt):
            raise ValueError(f"The value {t} is out of bounds. It must be between {self.mint} and {self.maxt}")
        else:
            return (self.o + self.d*t)
        
    def __str__(self):
        if self.maxt == float("inf"):
            return f'({self.o.x}, {self.o.y}, {self.o.z}) + t*({self.d.x}, {self.d.y}, {self.d.z})' + \
                f' with t ∈ [{self.mint},{self.maxt}['
        else:
            return f'({self.o.x}, {self.o.y}, {self.o.z}) + t*({self.d.x}, {self.d.y}, {self.d.z})' + \
                f' with t ∈ [{self.mint},{self.maxt}]'
        
    def __repr__(self):
        if self.maxt == float("inf"):
            return f'r(t) = ({self.o.x}, {self.o.y}, {self.o.z}) + t*({self.d.x}, {self.d.y}, {self.d.z})' + \
                f' with t ∈ [{self.mint},{self.maxt}['
        else:
            return f'r(t) = ({self.o.x}, {self.o.y}, {self.o.z}) + t*({self.d.x}, {self.d.y}, {self.d.z})' + \
                f' with t ∈ [{self.mint},{self.maxt}]'


class BBox(object):
    '''
    Bounding Box

    Parameters
    ----------
    p1 : Point, optional
        Frist point to use to create the BBox
    p2 : Point, optional
        Second point to use to create the BBox

    Examples
    --------
    >>> import geoclide as gc
    >>> p1 = gc.Point(0., 0., 0.)
    >>> p2 = gc.Point(1., 1., 1.)
    >>> b1 = gc.BBox(p1, p2)
    >>> b1
    pmin=Point(0.0, 0.0, 0.0), pmax=Point(1.0, 1.0, 1.0)
    '''
    def __init__(self, p1=None, p2=None):
        if (isinstance(p1, Point)  and isinstance(p2, Point)):
            self.pmin = Point(min(p1.x, p2.x), min(p1.y, p2.y), min(p1.z, p2.z))
            self.pmax = Point(max(p1.x, p2.x), max(p1.y, p2.y), max(p1.z, p2.z))
        elif (p1 is None and p2 is None):
            self.pmin = Point(float("inf"), float("inf"), float("inf"))
            self.pmax = Point(float("-inf"), float("-inf"), float("-inf"))
        elif (isinstance(p1, Point) and p2 is None):
            self.pmin = p1
            self.pmax = p1
        elif (p1 is None and isinstance(p2, Point)):
            self.pmin = p2
            self.pmax = p2
        else:
            raise ValueError('The only parameters accepted are Point objects')
        
        # The 8 vertices of the BBox
        # - p0=pmin, then next 3 points are in the XY plane at z=pmin.z the order being anti-clockwise
        # - next 4 points are in the XY plane at z=pmax.z, starting with point p4 just above p0, so p6=pmax
        self.p0 = Point(self.pmin.x,self.pmin.y,self.pmin.z)
        self.p1 = Point(self.pmax.x,self.pmin.y,self.pmin.z)
        self.p2 = Point(self.pmax.x,self.pmax.y,self.pmin.z)
        self.p3 = Point(self.pmin.x,self.pmax.y,self.pmin.z)
        self.p4 = Point(self.pmin.x,self.pmin.y,self.pmax.z)
        self.p5 = Point(self.pmax.x,self.pmin.y,self.pmax.z)
        self.p6 = Point(self.pmax.x,self.pmax.y,self.pmax.z)
        self.p7 = Point(self.pmin.x,self.pmax.y,self.pmax.z)
        self.vertices = [self.p0, self.p1, self.p2, self.p3,
                         self.p4, self.p5, self.p6, self.p7]
        
    def __str__(self):
        return f'pmin=({self.pmin.x}, {self.pmin.y}, {self.pmin.z}), pmax=({self.pmax.x}, {self.pmax.y}, {self.pmax.z})'
        
    def __repr__(self):
        return f'pmin=Point({self.pmin.x}, {self.pmin.y}, {self.pmin.z}), pmax=Point({self.pmax.x}, {self.pmax.y}, {self.pmax.z})'
        
    def union(self, b):
        """
        Union with a Point or another BBox

        Parameters
        ----------
        b : Point | BBox
            The point or BBox to use for the union

        Returns
        -------
        b_union : BBox
            The new BBox after the union
        
        Examples
        --------
        >>> import geoclide as gc
        >>> p1 = gc.Point(0., 0., 0.)
        >>> p2 = gc.Point(1., 1., 1.)
        >>> p3 = gc.Point(1., 1., 3.)
        >>> b1 = gc.BBox(p1, p2)
        >>> b1
        pmin=Point(0.0, 0.0, 0.0), pmax=Point(1.0, 1.0, 1.0)
        >>> b2 = b1.union(p3)
        >>> b2
        pmin=Point(0.0, 0.0, 0.0), pmax=Point(1.0, 1.0, 3.0)
        """
        b_union = BBox()
        if isinstance(b, Point):
            b_union.pmin.x = min(self.pmin.x, b.x)
            b_union.pmin.y = min(self.pmin.y, b.y)
            b_union.pmin.z = min(self.pmin.z, b.z)
            b_union.pmax.x = max(self.pmax.x, b.x)
            b_union.pmax.y = max(self.pmax.y, b.y)
            b_union.pmax.z = max(self.pmax.z, b.z)
        elif isinstance(b, BBox):
            b_union.pmin.x = min(self.pmin.x, b.pmin.x)
            b_union.pmin.y = min(self.pmin.y, b.pmin.y)
            b_union.pmin.z = min(self.pmin.z, b.pmin.z)
            b_union.pmax.x = max(self.pmax.x, b.pmax.x)
            b_union.pmax.y = max(self.pmax.y, b.pmax.y)
            b_union.pmax.z = max(self.pmax.z, b.pmax.z)
        else:
            raise ValueError('The union must be with another BBox or Point')

        return b_union

    def is_inside(self, P):
        """
        Test if Point P is included in BBox
        """
        return (P.x >= self.pmin.x) and (P.x <= self.pmax.x) and \
               (P.y >= self.pmin.y) and (P.y <= self.pmax.y) and \
               (P.z >= self.pmin.z) and (P.z <= self.pmax.z)

    def is_intersection(self, r1) :
        """
        Test if a ray intersects the BBox

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test

        Returns
        -------
        out : bool,
            If there is at least 1 intersection -> True, else False.

        Examples
        --------
        >>> import geoclide as gc
        >>> p1 = gc.Point(0., 0., 0.)
        >>> p2 = gc.Point(1., 1., 1.)
        >>> b1 = gc.BBox(p1, p2)
        pmin=Point(0.0, 0.0, 0.0), pmax=Point(1.0, 1.0, 1.0)
        >>> p3 = gc.Point(0.5, 0.5, 0.1)
        >>> v1 = gc.Vector(0., 0., 1.)
        >>> r1 = gc.Ray(p3, v1)
        >>> r1
        r(t) = (0.5, 0.5, 0.1) + t*(0.0, 0.0, 1.0) with t ∈ [0,inf[
        >>> b1.is_intersection(r1)
        True
        """
        if not isinstance(r1, Ray): raise ValueError('The given parameter must be a Ray')
        t0 = 0.
        t1 = r1.maxt
        for i in range(3):
            if r1.d[i]!= 0 : invRayDir = 1. / r1.d[i]
            else : invRayDir = math.inf
            tNear = (self.pmin[i] - r1.o[i]) * invRayDir
            tFar  = (self.pmax[i] - r1.o[i]) * invRayDir
            if (tNear > tFar): tNear, tFar = tFar, tNear
            tFar *= 1 + 2*GAMMA3_F64
            t0 = tNear if tNear > t0 else t0
            t1 = tFar  if  tFar < t1 else t1
            if (t0 > t1) : return False
        return True

    def intersect(self, r1) :
        """
        Test if a ray intersects the BBox

        There are 3 possibilities:
        - no intersection
        - only 1 intersection (case of ray located initially inside the BBox)
        - 2 intersections

        Parameters
        ----------
        r1 : Ray
            The ray to use for the intersection test

        Returns
        -------
        t0 : float
            The t ray variable for the first intersection.
            In case of only 1 intersection it represents nothing. 
        t1 : float
            The t ray variable for the second intersection.
            In case of only 1 intersection, t1 becomes the t ray variable for the first intersection.
        is_intersection : bool
            If there is at least 1 intersection -> True, else False.

        Examples
        --------
        >>> import geoclide as gc
        >>> p1 = gc.Point(0., 0., 0.)
        >>> p2 = gc.Point(1., 1., 1.)
        >>> b1 = gc.BBox(p1, p2)
        pmin=Point(0.0, 0.0, 0.0), pmax=Point(1.0, 1.0, 1.0)
        >>> p3 = gc.Point(0.5, 0.5, 0.1)
        >>> v1 = gc.Vector(0., 0., 1.)
        >>> r1 = gc.Ray(p3, v1)
        >>> r1
        r(t) = (0.5, 0.5, 0.1) + t*(0.0, 0.0, 1.0) with t ∈ [0,inf[
        >>> t0, t1, is_intersection = b1.intersect(r1)
        >>> t0, t1, is_intersection
        (0.0, 0.9, True)
        >>> r1[t1]
        Point(0.5, 0.5, 1.0)
        """
        if not isinstance(r1, Ray): raise ValueError('The given parameter must be a Ray')
        t0 = 0.
        t1 = r1.maxt
        for i in range(3):
            if r1.d[i]!= 0 : invRayDir = 1. / r1.d[i]
            else : invRayDir = math.inf
            tNear = (self.pmin[i] - r1.o[i]) * invRayDir
            tFar  = (self.pmax[i] - r1.o[i]) * invRayDir
            if (tNear > tFar): tNear, tFar = tFar, tNear
            tFar *= 1 + 2*GAMMA3_F64
            t0 = tNear if tNear > t0 else t0
            t1 = tFar  if  tFar < t1 else t1
            if (t0 > t1) : return 0., 0., False
        return t0, t1, True

    def common_vertices(self, b):
        """
        Return a list of boolean checking which vertices (self) are common to
        the BBox b

        Parameters
        ----------
        b : BBox
            The secondary Bounding Box
        
        Returns
        -------
        out : np.ndarray
        Return an array of boolean indicating if the BBox vertices are
        common to the secondary BBox (b) vertices

        Examples
        --------
        >>> import geoclide as gc
        >>> b0 = gc.BBox(gc.Point(0., 0., 0.), gc.Point(1., 1., 1.))
        >>> b1 = gc.BBox(gc.Point(1., 0., 0.), gc.Point(2., 1., 1.))
        >>> b0.common_vertices(b1)
        array([False,  True,  True, False, False,  True,  True, False])
        >>> b1.common_vertices(b0)
        array([ True, False, False,  True,  True, False, False,  True])
        """
        return get_common_vertices(self,b)

    def common_face(self, b, fill_value=None):
        """
        Return the face index which is common with one of the face of BBox b2
    
        The convention of index from face 0 to 5, for +X,-X,+Y,-Y,+Z,-Z:

        >>>    |F2|                     |+Y|
        >>> |F1|F4|F0|F5|  where ->  |-X|+Z|+X|-Z|
        >>>    |F3|                     |-Y|

        More information see: `en.wikipedia.org/wiki/Cube_mapping`

        Parameters
        ----------
        b : BBox
            The secondary Bounding Box
        fill_value : integer, optional
            In case there is no common face return fill_value

        Returns
        -------
        out : integer | None
            Return the index of the common face or fill_value
        
        Examples
        --------
        >>> import geoclide as gc
        >>> b0 = gc.BBox(gc.Point(0., 0., 0.), gc.Point(1., 1., 1.))
        >>> b1 = gc.BBox(gc.Point(1., 0., 0.), gc.Point(2., 1., 1.))
        >>> gc.get_common_face(b1, b2)
        0
        >>> gc.get_common_face(b2, b1)
        1
        """
        return get_common_face(self, b, fill_value=fill_value)


def get_common_vertices(b1, b2):
    """
    Check which vertices of BBox b1 are common to the vectices of BBox b2

    Parameters
    ----------
    b1 : BBox
        The principal Bounding Box
    b2 : BBox
        The secondary Bounding Box
    
    Returns
    -------
    out : np.ndarray
        Return an array of boolean indicating if the principal BBox (b1)
        vertices are common to secondary BBox (b2) vertices

    Examples
    --------
    >>> import geoclide as gc
    >>> b0 = gc.BBox(gc.Point(0., 0., 0.), gc.Point(1., 1., 1.))
    >>> b1 = gc.BBox(gc.Point(1., 0., 0.), gc.Point(2., 1., 1.))
    >>> gc.get_common_vertices(b1, b2)
    array([False,  True,  True, False, False,  True,  True, False])
    >>> gc.get_common_vertices(b1, b2)
    array([ True, False, False,  True,  True, False, False,  True])
    """
    return np.array(list((map(lambda x: x in b2.vertices, b1.vertices))))


def get_common_face(b1, b2, fill_value=None):
    """

    Return the face index of the BBox b1 which is common to BBox b2
    
    The convention of index from face 0 to 5, for +X,-X,+Y,-Y,+Z,-Z:

    >>>    |F2|                     |+Y|
    >>> |F1|F4|F0|F5|  where ->  |-X|+Z|+X|-Z|
    >>>    |F3|                     |-Y|

    More information see: `en.wikipedia.org/wiki/Cube_mapping`

    Parameters
    ----------
    b1 : BBox
        The principal Bounding Box
    b2 : BBox
        The secondary Bounding Box
    fill_value : integer, optional
            In case there is no common face return fill_value

    Returns
    -------
    out : integer | None
        Return the index of the common face or fill_value

    Examples
    --------
    >>> import geoclide as gc
    >>> b0 = gc.BBox(gc.Point(0., 0., 0.), gc.Point(1., 1., 1.))
    >>> b1 = gc.BBox(gc.Point(1., 0., 0.), gc.Point(2., 1., 1.))
    >>> gc.get_common_face(b1, b2)
    0
    >>> gc.get_common_face(b2, b1)
    1
    """
    ok = get_common_vertices(b1,b2)
    if ok.sum()==4:
        n  = np.arange(8)[ok]
        if   np.array_equal(n, np.array([1,2,5,6])):
            return 0
        elif np.array_equal(n, np.array([0,3,4,7])):
            return 1
        elif np.array_equal(n, np.array([2,3,6,7])):
            return 2
        elif np.array_equal(n, np.array([0,1,4,5])):
            return 3
        elif np.array_equal(n, np.array([4,5,6,7])):
            return 4
        elif np.array_equal(n, np.array([0,1,2,3])):
            return 5
        else: return fill_value

    else :
        return fill_value